import datetime
import logging
from itertools import chain

import ipywidgets as widgets
from IPython.display import display

logger = logging.getLogger(__name__)


class ExpectationExplorer:
    def __init__(self):
        self.state = {"data_assets": {}}
        self.expectation_kwarg_field_names = {
            "expect_column_values_to_be_unique": ["mostly"],
            "expect_column_unique_value_count_to_be_between": [
                "min_value",
                "max_value",
            ],
            "expect_column_value_lengths_to_be_between": [
                "min_value",
                "max_value",
                "mostly",
            ],
            "expect_table_row_count_to_be_between": ["min_value", "max_value"],
            "expect_table_column_count_to_be_between": ["min_value", "max_value"],
            "expect_column_proportion_of_unique_values_to_be_between": [
                "min_value",
                "max_value",
            ],
            "expect_column_median_to_be_between": ["min_value", "max_value"],
            "expect_column_mean_to_be_between": ["min_value", "max_value"],
            "expect_column_stdev_to_be_between": ["min_value", "max_value"],
            "expect_column_quantile_values_to_be_between": ["quantile_ranges"],
            "expect_column_sum_to_be_between": ["min_value", "max_value"],
            "expect_column_values_to_not_be_null": ["mostly"],
            "expect_column_values_to_be_null": ["mostly"],
            "expect_column_values_to_be_json_parseable": ["mostly"],
            "expect_column_values_to_be_dateutil_parseable": ["mostly"],
            "expect_column_values_to_be_increasing": [
                "strictly",
                "parse_strings_as_datetimes",
                "mostly",
            ],
            "expect_column_values_to_be_decreasing": [
                "strictly",
                "parse_strings_as_datetimes",
                "mostly",
            ],
            "expect_column_values_to_match_regex": ["regex", "mostly"],
            "expect_column_values_to_not_match_regex": ["regex", "mostly"],
            "expect_column_values_to_match_json_schema": ["json_schema", "mostly"],
            "expect_column_values_to_match_regex_list": [
                "regex_list",
                "match_on",
                "mostly",
            ],
            "expect_column_values_to_not_match_regex_list": ["regex_list", "mostly"],
            "expect_column_values_to_be_in_set": [
                "value_set",
                "parse_strings_as_datetimes",
                "mostly",
            ],
            "expect_column_values_to_not_be_in_set": [
                "value_set",
                "parse_strings_as_datetimes",
                "mostly",
            ],
            "expect_column_distinct_values_to_be_in_set": [
                "value_set",
                "parse_strings_as_datetimes",
            ],
            "expect_column_distinct_values_to_equal_set": [
                "value_set",
                "parse_strings_as_datetimes",
            ],
            "expect_column_distinct_values_to_contain_set": [
                "value_set",
                "parse_strings_as_datetimes",
            ],
            "expect_column_most_common_value_to_be_in_set": ["value_set", "ties_okay"],
            "expect_column_to_exist": ["column_index"],
            "expect_column_value_lengths_to_equal": ["value", "mostly"],
            "expect_table_row_count_to_equal": ["value"],
            "expect_table_column_count_to_equal": ["value"],
            "expect_column_values_to_match_strftime_format": [
                "strftime_format",
                "mostly",
            ],
            "expect_column_values_to_be_between": [
                "min_max_type",
                "parse_strings_as_datetimes",
                "output_strftime_format",
                "min_value",
                "max_value",
                "mostly",
                "allow_cross_type_comparisons",
            ],
            "expect_column_max_to_be_between": [
                "min_max_type",
                "parse_strings_as_datetimes",
                "output_strftime_format",
                "min_value",
                "max_value",
            ],
            "expect_column_min_to_be_between": [
                "min_max_type",
                "parse_strings_as_datetimes",
                "output_strftime_format",
                "min_value",
                "max_value",
            ],
            "expect_table_columns_to_match_ordered_list": ["column_list"],
            "expect_table_columns_to_match_set": ["column_set", "exact_match"],
            ####
            "expect_column_pair_values_to_be_equal": ["ignore_row_if"],
            "expect_column_pair_values_A_to_be_greater_than_B": [
                "or_equal",
                "ignore_row_if",
                "allow_cross_type_comparisons",
            ],
            "expect_column_pair_values_to_be_in_set": [
                "value_pairs_set",
                "ignore_row_if",
            ],
            "expect_compound_columns_to_be_unique": ["ignore_row_if"],
            "expect_select_column_values_to_be_unique_within_record": ["ignore_row_if"],
            "expect_column_values_to_be_of_type": ["type_", "mostly"],
            "expect_column_values_to_be_in_type_list": ["type_list", "mostly"],
            "expect_column_kl_divergence_to_be_less_than": [
                "partition_object",
                "threshold",
                "internal_weight_holdout",
                "tail_weight_holdout",
            ],
            "expect_column_chisquare_test_p_value_to_be_greater_than": [
                "partition_object",
                "p",
            ],
            "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than": [
                "partition_object",
                "p",
                "bootstrap_samples",
                "bootstrap_sample_size",
            ],
            "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than": [
                "distribution",
                "p_value",
                "params",
            ],
            "_expect_column_values_to_be_in_type_list__aggregate": ["type_list"],
            "_expect_column_values_to_be_in_type_list__map": ["type_list"],
        }
        self.kwarg_widget_exclusions = ["column", "result_format", "include_config"]
        self.min_max_subtypes = {
            "integer": {
                "positive": [
                    "expect_column_unique_value_count_to_be_between",
                    "expect_column_value_lengths_to_be_between",
                    "expect_table_row_count_to_be_between",
                ],
                "unbounded": ["expect_column_median_to_be_between"],
            },
            "float": {
                "unit_interval": [
                    "expect_column_proportion_of_unique_values_to_be_between"
                ],
                "unbounded": [
                    "expect_column_mean_to_be_between",
                    "expect_column_stdev_to_be_between",
                    "expect_column_sum_to_be_between",
                ],
            },
            "mixed": {
                "unbounded": [
                    "expect_column_values_to_be_between",
                    "expect_column_min_to_be_between",
                    "expect_column_max_to_be_between",
                ]
            },
        }
        self.debug_view = widgets.Output()
        self.styles = {"description_width": {"description_width": "150px"}}

    def update_result(self, data_asset_name, new_result, column=None):
        new_success_value = new_result.get("success")
        expectation_type = new_result.expectation_config.expectation_type
        new_result_widgets = self.generate_expectation_result_detail_widgets(
            result=new_result.result
        )
        new_border_color = "green" if new_success_value else "red"
        data_asset_expectations = self.state["data_assets"][data_asset_name][
            "expectations"
        ]

        if column:
            data_asset_expectations[column][expectation_type][
                "success"
            ].value = "<span><strong>Success: </strong>{new_success_value}</span>".format(
                new_success_value=str(new_success_value)
            )
            data_asset_expectations[column][expectation_type][
                "result_detail_widget"
            ].children = new_result_widgets
            data_asset_expectations[column][expectation_type][
                "editor_widget"
            ].layout.border = "2px solid {new_border_color}".format(
                new_border_color=new_border_color
            )
        else:
            data_asset_expectations["non_column_expectations"][expectation_type][
                "success"
            ].value = "<span><strong>Success: </strong>{new_success_value}</span>".format(
                new_success_value=str(new_success_value)
            )
            data_asset_expectations["non_column_expectations"][expectation_type][
                "result_detail_widget"
            ].children = new_result_widgets
            data_asset_expectations["non_column_expectations"][expectation_type][
                "editor_widget"
            ].layout.border = "2px solid {new_border_color}".format(
                new_border_color=new_border_color
            )

    def get_expectation_state(self, data_asset_name, expectation_type, column=None):
        data_asset_state = self.state["data_assets"].get(data_asset_name, {})
        data_asset_expectations = data_asset_state.get("expectations", {})

        if column:
            column_expectations = data_asset_expectations.get(column)
            if not column_expectations:
                return None
            return column_expectations.get(expectation_type)
        else:
            non_column_expectations = data_asset_expectations.get(
                "non_column_expectations"
            )
            if not non_column_expectations:
                return None
            return non_column_expectations.get(expectation_type)

    def initialize_data_asset_state(self, data_asset):
        data_asset_name = data_asset.data_asset_name

        self.state["data_assets"][data_asset_name] = {
            "data_asset": data_asset,
            "expectations": {},
        }

    def set_expectation_state(self, data_asset, expectation_state, column=None):
        data_asset_name = data_asset.data_asset_name
        expectation_type = expectation_state.get("expectation_type")
        data_asset_state = self.state["data_assets"].get(data_asset_name)

        data_asset_expectations = data_asset_state.get("expectations")

        if column:
            column_expectations = data_asset_expectations.get(column, {})
            column_expectations[expectation_type] = expectation_state
            data_asset_expectations[column] = column_expectations
        else:
            non_column_expectations = data_asset_expectations.get(
                "non_column_expectations", {}
            )
            non_column_expectations[expectation_type] = expectation_state
            data_asset_expectations["non_column_expectations"] = non_column_expectations

    # interconvert expectation kwargs
    def expectation_kwarg_dict_to_ge_kwargs(self, kwarg_dict):
        # kwarg_exclusions are non-ge keys which contain ancillary widgets required for proper parsing
        # for e.g., a widget for setting the date type of min/max
        ancillary_kwargs = ["min_max_type"]
        expectation_kwargs = {}

        for kwarg_name, widget_dict in kwarg_dict.items():
            if kwarg_name == "column":
                expectation_kwargs["column"] = widget_dict
                continue
            if kwarg_name in ancillary_kwargs:
                continue
            elif not hasattr(
                self, "generate_{kwarg_name}_widget_dict".format(kwarg_name=kwarg_name)
            ):
                expectation_kwargs[kwarg_name] = widget_dict.get("ge_kwarg_value")
            else:
                expectation_kwargs[kwarg_name] = (
                    widget_dict.get("ge_kwarg_value")
                    if "ge_kwarg_value" in widget_dict
                    else widget_dict["kwarg_widget"].value
                )

        return expectation_kwargs

    def update_kwarg_widget_dict(
        self, expectation_state, current_widget_dict, ge_kwarg_name, new_ge_kwarg_value
    ):
        def update_tag_list_widget_dict(widget_dict, new_list):
            widget_dict["ge_kwarg_value"] = new_list
            widget_display = widget_dict["widget_display"]
            widget_display.children = self.generate_tag_button_list(
                expectation_state=expectation_state,
                tag_list=new_list,
                widget_display=widget_display,
            )

        def ge_number_to_widget_string(widget_dict, number_kwarg):
            if hasattr(widget_dict["kwarg_widget"], "value"):
                widget_dict["kwarg_widget"].value = (
                    str(number_kwarg) if number_kwarg or number_kwarg == 0 else ""
                )
            if widget_dict.get("ge_kwarg_value"):
                widget_dict["ge_kwarg_value"] = (
                    str(number_kwarg) if number_kwarg or number_kwarg == 0 else ""
                )

        def min_max_value_to_string(widget_dict, number_kwarg):
            # expectations with min/max kwargs where widget expects a string
            number_to_string_expectations = [
                "expect_column_values_to_be_between",
                "expect_column_max_to_be_between",
                "expect_column_min_to_be_between",
            ]

            if (
                expectation_state.get("expectation_type")
                in number_to_string_expectations
            ):
                ge_number_to_widget_string(widget_dict, number_kwarg)

        updater_functions = {
            "regex_list": update_tag_list_widget_dict,
            "value_set": update_tag_list_widget_dict,
            "column_list": update_tag_list_widget_dict,
            "column_index": ge_number_to_widget_string,
            "min_value": min_max_value_to_string,
            "max_value": min_max_value_to_string,
        }
        updater_functions.get(
            ge_kwarg_name,
            lambda current_widget_dict, ge_kwarg_value: setattr(
                current_widget_dict["kwarg_widget"], "value", ge_kwarg_value
            ),
        )(current_widget_dict, new_ge_kwarg_value)

    def update_expectation_state(
        self, existing_expectation_state, expectation_validation_result, validation_time
    ):
        expectation_editor_widget = existing_expectation_state.get("editor_widget")
        new_ge_expectation_kwargs = expectation_validation_result.expectation_config[
            "kwargs"
        ]
        current_expectation_kwarg_dict = existing_expectation_state["kwargs"]
        column = current_expectation_kwarg_dict.get("column")
        data_asset_name = existing_expectation_state.get("data_asset_name")

        for ge_kwarg_name, ge_kwarg_value in new_ge_expectation_kwargs.items():
            if ge_kwarg_name in self.kwarg_widget_exclusions:
                continue

            current_widget_dict = current_expectation_kwarg_dict.get(ge_kwarg_name)

            if current_widget_dict:
                if not hasattr(
                    self,
                    "generate_{ge_kwarg_name}_widget_dict".format(
                        ge_kwarg_name=ge_kwarg_name
                    ),
                ):
                    current_expectation_kwarg_dict[
                        ge_kwarg_name
                    ] = self.generate_expectation_kwarg_fallback_widget_dict(
                        expectation_kwarg_name=ge_kwarg_name,
                        **new_ge_expectation_kwargs
                    )
                else:
                    self.update_kwarg_widget_dict(
                        expectation_state=existing_expectation_state,
                        current_widget_dict=current_widget_dict,
                        ge_kwarg_name=ge_kwarg_name,
                        new_ge_kwarg_value=ge_kwarg_value,
                    )
            else:
                widget_dict_generator = getattr(
                    self,
                    "generate_{ge_kwarg_name}_widget_dict".format(
                        ge_kwarg_name=ge_kwarg_name
                    ),
                    None,
                )
                widget_dict = (
                    widget_dict_generator(
                        expectation_state=existing_expectation_state,
                        **new_ge_expectation_kwargs
                    )
                    if widget_dict_generator
                    else self.generate_expectation_kwarg_fallback_widget_dict(
                        expectation_kwarg_name=ge_kwarg_name,
                        **new_ge_expectation_kwargs
                    )
                )
                current_expectation_kwarg_dict[ge_kwarg_name] = widget_dict
                expectation_editor_widget.children[0].children[1].children += (
                    widget_dict["kwarg_widget"],
                )

        self.update_result(
            data_asset_name=data_asset_name,
            new_result=expectation_validation_result,
            column=column,
        )
        # TODO: This is messy. Figure out better way of storing/accessing UI elements
        validation_time_widget = (
            expectation_editor_widget.children[0].children[0].children[0].children[-1]
        )
        validation_time_widget.value = "<div><strong>Date/Time Validated (UTC): </strong>{validation_time}</div>".format(
            validation_time=validation_time
        )
        return expectation_editor_widget

    # widget generators for general input fields
    def generate_boolean_checkbox_widget(
        self, value, description="", description_tooltip="", disabled=False
    ):
        return widgets.Checkbox(
            value=value,
            description=description,
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            description_tooltip=description_tooltip,
            disabled=disabled,
        )

    def generate_text_area_widget(
        self,
        value,
        description="",
        description_tooltip="",
        continuous_update=False,
        placeholder="",
    ):
        return widgets.Textarea(
            value=value,
            placeholder=placeholder,
            description="<strong>{description}: </strong>".format(
                description=description
            ),
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            description_tooltip=description_tooltip,
            continuous_update=continuous_update,
        )

    def generate_text_widget(
        self,
        value,
        description="",
        description_tooltip="",
        continuous_update=False,
        placeholder="",
        disabled=False,
    ):
        return widgets.Text(
            value=value,
            placeholder=placeholder,
            description="<strong>{description}: </strong>".format(
                description=description
            ),
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            description_tooltip=description_tooltip,
            continuous_update=continuous_update,
            disabled=disabled,
        )

    def generate_radio_buttons_widget(
        self, value, options=None, description="", description_tooltip=""
    ):
        if options is None:
            options = []

        return widgets.RadioButtons(
            options=options,
            value=value,
            description="<strong>{description}: </strong>".format(
                description=description
            ),
            layout={"width": "400px"},
            style=self.styles.get("description_width"),
            description_tooltip=description_tooltip,
        )

    def generate_remove_expectation_button(self, expectation_state):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        column = expectation_state["kwargs"].get("column")

        remove_expectation_button = widgets.Button(
            description="Remove Expectation",
            button_style="danger",
            tooltip="click to remove expectation",
            icon="trash",
            layout={"width": "auto"},
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_click(button):
            editor_widget = expectation_state.get("editor_widget")
            expectation = data_asset.remove_expectation(
                expectation_type=expectation_type, column=column
            )
            data_asset_state = self.state["data_assets"].get(data_asset_name, {})
            data_asset_expectations = data_asset_state.get("expectations")

            if column:
                column_expectations = data_asset_expectations.get(column)
                column_expectations.pop(expectation_type)
            else:
                non_column_expectations = data_asset_expectations.get(
                    "non_column_expectations"
                )
                non_column_expectations.pop(expectation_type)

            editor_widget.close()

        remove_expectation_button.on_click(on_click)

        return remove_expectation_button

    def generate_tag_button(self, expectation_state, tag, tag_list, widget_display):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        tag_button = widgets.Button(
            description=str(tag),
            button_style="info",
            tooltip="click to delete",
            icon="trash",
            layout={"width": "auto"},
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_click(button):
            if len(tag_list) == 1:
                with expectation_feedback_widget:
                    print("Warning: cannot remove - set must have at least one member")
                return
            tag_list.remove(tag)

            display_children = list(widget_display.children)
            display_children.remove(tag_button)
            widget_display.children = display_children

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        tag_button.on_click(on_click)

        return tag_button

    def generate_tag_button_list(self, expectation_state, tag_list, widget_display):
        return (
            [
                self.generate_tag_button(
                    expectation_state=expectation_state,
                    tag=tag,
                    tag_list=tag_list,
                    widget_display=widget_display,
                )
                for tag in tag_list
            ]
            if tag_list
            else []
        )

    def generate_zero_or_positive_integer_widget(
        self, value, max=int(9e300), description="", continuous_update=False
    ):
        return widgets.BoundedIntText(
            value=value,
            min=0,
            max=max,
            description="<strong>{description}: </strong>".format(
                description=description
            ),
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            continuous_update=continuous_update,
        )

    # widget dict generators for kwarg input fields
    def generate_output_strftime_format_widget_dict(
        self,
        expectation_state,
        output_strftime_format="",
        column=None,
        **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        parse_strings_as_datetimes = expectation_kwargs.get(
            "parse_strings_as_datetimes", False
        )

        output_strftime_format_widget = self.generate_text_widget(
            value=output_strftime_format,
            description="output strftime format",
            placeholder="press enter to confirm...",
            description_tooltip="Enter a valid strfime format for datetime output.",
            disabled=not parse_strings_as_datetimes,
        )

        widget_dict = {"kwarg_widget": output_strftime_format_widget}

        @expectation_feedback_widget.capture(clear_output=True)
        def on_output_strftime_format_submit(widget):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        output_strftime_format_widget.on_submit(on_output_strftime_format_submit)

        return widget_dict

    def generate_strftime_format_widget_dict(
        self, expectation_state, strftime_format="", column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        strftime_format_widget = self.generate_text_widget(
            value=strftime_format,
            description="strftime format",
            placeholder="press enter to confirm...",
        )

        widget_dict = {"kwarg_widget": strftime_format_widget}

        @expectation_feedback_widget.capture(clear_output=True)
        def on_strftime_format_submit(widget):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        strftime_format_widget.on_submit(on_strftime_format_submit)

        return widget_dict

    def generate_value_widget_dict(
        self, expectation_state, value=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        value_widget = self.generate_zero_or_positive_integer_widget(
            value=value, description="value"
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_value_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        value_widget.observe(on_value_change, names="value")

        widget_dict = {"kwarg_widget": value_widget}

        return widget_dict

    def generate_json_schema_widget_dict(
        self, expectation_state, json_schema="", column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        json_schema_widget = self.generate_text_area_widget(
            value=json_schema,
            description="json schema",
            placeholder="enter json schema",
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_json_schema_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        json_schema_widget.observe(on_json_schema_change, names="value")

        widget_dict = {"kwarg_widget": json_schema_widget}

        return widget_dict

    def generate_ties_okay_widget_dict(
        self, expectation_state, ties_okay=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        ties_okay_widget = self.generate_boolean_checkbox_widget(
            value=ties_okay,
            description="ties_okay",
            description_tooltip="If True, then the expectation will still succeed if values outside the designated set are as common (but not more common) than designated values.",
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_ties_okay_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        ties_okay_widget.observe(on_ties_okay_change, names="value")

        widget_dict = {"kwarg_widget": ties_okay_widget}

        return widget_dict

    def generate_match_on_widget_dict(
        self, expectation_state, match_on="any", column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        match_on_widget = self.generate_radio_buttons_widget(
            value=match_on,
            options=["any", "all"],
            description="match_on",
            description_tooltip="Use “any” if the value should match at least one regular expression in the list. Use “all” if it should match each regular expression in the list.",
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_match_on_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        match_on_widget.observe(on_match_on_change, names="value")

        widget_dict = {"kwarg_widget": match_on_widget}

        return widget_dict

    def generate_regex_list_widget_dict(
        self, expectation_state, regex_list=None, column=None, **expectation_kwargs
    ):
        if regex_list is None:
            regex_list = []

        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        regex_list_widget_display = widgets.Box(
            layout={"display": "flex", "flex_flow": "row wrap"}
        )
        regex_list_widget_input = self.generate_text_widget(
            value=None,
            description="add regex",
            placeholder="press enter to add...",
            continuous_update=True,
        )
        regex_list_label = widgets.HTML(
            value='<div style="text-align:right"><strong>regex list: </strong></div>',
            layout={"width": "150px"},
        )
        regex_list_widget = widgets.VBox(
            [
                widgets.HBox([regex_list_label, regex_list_widget_display]),
                regex_list_widget_input,
            ]
        )

        widget_dict = {
            "kwarg_widget": regex_list_widget,
            "ge_kwarg_value": regex_list,
            "widget_display": regex_list_widget_display,
            "widget_input": regex_list_widget_input,
        }

        regex_list_widget_display.children = self.generate_tag_button_list(
            expectation_state=expectation_state,
            tag_list=regex_list,
            widget_display=regex_list_widget_display,
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_regex_list_input_submit(widget):
            widget_display = expectation_state["kwargs"]["regex_list"]["widget_display"]
            regex_list = expectation_state["kwargs"]["regex_list"]["ge_kwarg_value"]
            new_regex = widget.value

            if new_regex in regex_list:
                with expectation_feedback_widget:
                    print("Warning: regex already exists in set")
                return

            regex_list.append(new_regex)
            widget_display.children += (
                self.generate_tag_button(
                    tag=new_regex,
                    expectation_state=expectation_state,
                    tag_list=regex_list,
                    widget_display=widget_display,
                ),
            )
            widget.value = ""

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        regex_list_widget_input.on_submit(on_regex_list_input_submit)

        return widget_dict

    def generate_column_list_widget_dict(
        self, expectation_state, column_list=None, column=None, **expectation_kwargs
    ):
        if column_list is None:
            column_list = []

        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        column_list_widget_display = widgets.Box(
            layout={"display": "flex", "flex_flow": "row wrap"}
        )
        column_list_widget_input = self.generate_text_widget(
            value=None,
            description="add column",
            placeholder="press enter to add...",
            continuous_update=True,
        )
        column_list_label = widgets.HTML(
            value='<div style="text-align:right"><strong>column list: </strong></div>',
            layout={"width": "150px"},
        )
        column_list_widget = widgets.VBox(
            [
                widgets.HBox([column_list_label, column_list_widget_display]),
                column_list_widget_input,
            ]
        )

        widget_dict = {
            "kwarg_widget": column_list_widget,
            "ge_kwarg_value": column_list,
            "widget_display": column_list_widget_display,
            "widget_input": column_list_widget_input,
        }

        column_list_widget_display.children = self.generate_tag_button_list(
            expectation_state=expectation_state,
            tag_list=column_list,
            widget_display=column_list_widget_display,
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_column_list_input_submit(widget):
            widget_display = expectation_state["kwargs"]["column_list"][
                "widget_display"
            ]
            column_list = expectation_state["kwargs"]["column_list"]["ge_kwarg_value"]
            new_column = widget.value

            if new_column in column_list:
                with expectation_feedback_widget:
                    print("Warning: column already exists in list")
                return

            column_list.append(new_column)
            widget_display.children += (
                self.generate_tag_button(
                    tag=new_column,
                    expectation_state=expectation_state,
                    tag_list=column_list,
                    widget_display=widget_display,
                ),
            )
            widget.value = ""

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        column_list_widget_input.on_submit(on_column_list_input_submit)

        return widget_dict

    def generate_column_index_widget_dict(
        self, expectation_state, column, column_index="", **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        column_index_widget = self.generate_text_widget(
            value=str(column_index),
            description="column_index",
            continuous_update=True,
            placeholder="press enter to confirm...",
        )

        widget_dict = {
            "kwarg_widget": column_index_widget,
            "ge_kwarg_value": column_index,
        }

        @expectation_feedback_widget.capture(clear_output=True)
        def on_column_index_submit(widget):
            new_column_index = widget.value if widget.value else None
            if new_column_index:
                if "." in new_column_index or "-" in new_column_index:
                    with expectation_feedback_widget:
                        print("Warning: column_index must be an integer >= 0")
                    return
                try:
                    new_column_index = int(new_column_index)
                except ValueError:
                    with expectation_feedback_widget:
                        print("Warning: column_index must be an integer >= 0")
                    return

            widget_dict["ge_kwarg_value"] = new_column_index

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        column_index_widget.on_submit(on_column_index_submit)

        return widget_dict

    def generate_regex_widget_dict(
        self, expectation_state, regex="", column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        regex_widget = self.generate_text_area_widget(
            value=regex, description="regex", placeholder=r"e.g. ([A-Z])\w+"
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_regex_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        regex_widget.observe(on_regex_change, names="value")

        widget_dict = {"kwarg_widget": regex_widget}

        return widget_dict

    def generate_parse_strings_as_datetimes_widget_dict(
        self,
        expectation_state,
        parse_strings_as_datetimes=None,
        column=None,
        **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        min_max_type_widget_dict = expectation_state["kwargs"].get("min_max_type", {})
        min_max_type_widget = min_max_type_widget_dict.get("kwarg_widget")
        min_max_type = min_max_type_widget.value if min_max_type_widget else None
        parse_strings_as_datetimes_widget = self.generate_boolean_checkbox_widget(
            value=parse_strings_as_datetimes,
            description="parse strings as datetimes",
            disabled=False if min_max_type == "string" or not min_max_type else True,
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_parse_strings_as_datetimes_change(change):
            output_strftime_format_widget_dict = expectation_state["kwargs"].get(
                "output_strftime_format", {}
            )
            output_strftime_format_widget = output_strftime_format_widget_dict.get(
                "kwarg_widget"
            )
            if output_strftime_format_widget:
                output_strftime_format_widget.disabled = not change["new"]

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        parse_strings_as_datetimes_widget.observe(
            on_parse_strings_as_datetimes_change, names="value"
        )

        widget_dict = {"kwarg_widget": parse_strings_as_datetimes_widget}

        return widget_dict

    def generate_strictly_widget_dict(
        self, expectation_state, strictly=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        inc_dec = expectation_type.split("_")[-1]
        strictly_widget = self.generate_boolean_checkbox_widget(
            value=strictly, description="strictly {inc_dec}".format(inc_dec=inc_dec)
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_strictly_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        strictly_widget.observe(on_strictly_change, names="value")

        widget_dict = {"kwarg_widget": strictly_widget}

        return widget_dict

    def generate_mostly_widget_dict(
        self, expectation_state, mostly=1, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        mostly_widget = widgets.FloatSlider(
            value=mostly,
            min=0,
            max=1.0,
            step=0.01,
            description="<strong>mostly: </strong>",
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_mostly_change(change):
            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        mostly_widget.observe(on_mostly_change, names="value")
        widget_dict = {"kwarg_widget": mostly_widget}

        return widget_dict

    def generate_min_max_type_widget_dict(
        self, expectation_state, type_option=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        min_value = expectation_kwargs.get("min_value")
        min_value_type = type(min_value)

        max_value = expectation_kwargs.get("max_value")
        max_value_type = type(max_value)

        number_types = [int, float]
        if min_value_type in number_types or max_value_type in number_types:
            type_option = "number"
        elif min_value_type is str or max_value_type is str:
            type_option = "string"

        min_max_type_widget = self.generate_radio_buttons_widget(
            value=type_option, options=["string", "number"], description="min/max type"
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_min_max_type_change(change):
            new_type_selection = change.get("new")

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )
            min_value = ge_expectation_kwargs.get("min_value")
            max_value = ge_expectation_kwargs.get("max_value")

            parse_strings_as_datetimes_widget_dict = expectation_state["kwargs"].get(
                "parse_strings_as_datetimes", {}
            )
            parse_strings_as_datetimes_widget = (
                parse_strings_as_datetimes_widget_dict.get("kwarg_widget")
            )

            if new_type_selection == "string":
                ge_expectation_kwargs["min_value"] = (
                    str(min_value) if min_value else None
                )
                ge_expectation_kwargs["max_value"] = (
                    str(max_value) if max_value else None
                )

                if parse_strings_as_datetimes_widget:
                    parse_strings_as_datetimes_widget.disabled = False
            elif new_type_selection == "number":
                ge_expectation_kwargs["min_value"] = (
                    float(min_value) if min_value else None
                )
                ge_expectation_kwargs["max_value"] = (
                    float(max_value) if max_value else None
                )

                if parse_strings_as_datetimes_widget:
                    parse_strings_as_datetimes_widget.value = False
                    parse_strings_as_datetimes_widget.disabled = True

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        min_max_type_widget.observe(on_min_max_type_change, names="value")

        return {"kwarg_widget": min_max_type_widget}

    def generate_min_value_widget_dict(
        self, expectation_state, min_value=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        min_value_widget_dict = expectation_state["kwargs"].get("min_value", {})
        min_value_widget = (
            min_value_widget_dict.get("kwarg_widget") if min_value_widget_dict else None
        )
        max_value_widget_dict = expectation_state["kwargs"].get("max_value", {})
        max_value_widget = (
            max_value_widget_dict.get("kwarg_widget") if max_value_widget_dict else None
        )

        # integer
        if expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["integer"].values())
        ):
            if expectation_type in self.min_max_subtypes["integer"]["positive"]:
                default_min_value = 0
                default_max_value = int(9e300)
            elif expectation_type in self.min_max_subtypes["integer"]["unbounded"]:
                default_min_value = -int(9e300)
                default_max_value = int(9e300)

            if min_value_widget:
                min_value_widget.value = min_value if min_value else default_min_value
            else:
                min_value_widget = widgets.BoundedIntText(
                    value=min_value or default_min_value,
                    min=default_min_value,
                    description="min_value: ",
                    disabled=False,
                )
            if not max_value_widget:
                max_value_widget = widgets.BoundedIntText(
                    description="max_value: ",
                    value=default_max_value,
                    max=default_max_value,
                    disabled=False,
                )
        # float
        elif expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["float"].values())
        ):
            if expectation_type in self.min_max_subtypes["float"]["unit_interval"]:
                default_min_value = 0
                default_max_value = 1
            elif expectation_type in self.min_max_subtypes["float"]["unbounded"]:
                default_min_value = -int(9e300)
                default_max_value = int(9e300)

            if min_value_widget:
                min_value_widget.value = min_value if min_value else default_min_value
            else:
                min_value_widget = widgets.BoundedFloatText(
                    value=min_value or default_min_value,
                    min=default_min_value,
                    step=0.01,
                    description="min_value: ",
                    disabled=False,
                )
            if not max_value_widget:
                max_value_widget = widgets.BoundedFloatText(
                    description="max_value: ",
                    value=default_max_value,
                    max=default_max_value,
                    step=0.01,
                    disabled=False,
                )
        # number or string
        elif expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["mixed"].values())
        ):
            min_value_str = str(min_value) if min_value or min_value == 0 else ""
            if min_value_widget:
                min_value_widget.value = min_value
            else:
                min_value_widget = self.generate_text_widget(
                    value=min_value_str,
                    description="min_value",
                    placeholder="press enter to confirm...",
                )

            @expectation_feedback_widget.capture(clear_output=True)
            def on_min_value_change(change):
                min_max_type = expectation_state["kwargs"]["min_max_type"][
                    "kwarg_widget"
                ].value
                ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                    expectation_state["kwargs"]
                )
                min_value = ge_expectation_kwargs.get("min_value")
                max_value = ge_expectation_kwargs.get("max_value")

                if min_max_type == "string":
                    min_value = str(min_value) if min_value else None
                    max_value = str(max_value) if max_value else None
                elif min_max_type == "number":
                    min_value = float(min_value) if min_value else None
                    max_value = float(max_value) if max_value else None

                if min_value and max_value and min_value > max_value:
                    with expectation_feedback_widget:
                        print(
                            "Error: min_value must be less than or equal to max_value."
                        )
                    return

                ge_expectation_kwargs["min_value"] = min_value
                ge_expectation_kwargs["max_value"] = max_value

                getattr(data_asset, expectation_type)(
                    include_config=True, **ge_expectation_kwargs
                )

            min_value_widget.observe(on_min_value_change, names="value")

        if (
            min_value_widget
            and max_value_widget
            and expectation_type
            not in list(chain.from_iterable(self.min_max_subtypes["mixed"].values()))
        ):

            @expectation_feedback_widget.capture(clear_output=True)
            def on_min_value_change(change):
                ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                    expectation_state["kwargs"]
                )
                getattr(data_asset, expectation_type)(
                    include_config=True, **ge_expectation_kwargs
                )

            min_value_widget.observe(on_min_value_change, names="value")
            max_dl = widgets.link(
                (max_value_widget, "value"), (min_value_widget, "max")
            )
            expectation_state["kwargs"]["max_value"] = {
                "kwarg_widget": max_value_widget
            }

        if min_value_widget:
            min_value_widget_dict["kwarg_widget"] = min_value_widget
        else:
            min_value_widget_dict = (
                self.generate_expectation_kwarg_fallback_widget_dict(
                    expectation_kwarg_name="min_value", **{"min_value": min_value}
                )
            )

        expectation_state["kwargs"]["min_value"] = min_value_widget_dict

        return min_value_widget_dict

    def generate_max_value_widget_dict(
        self, expectation_state, max_value=None, column=None, **expectation_kwargs
    ):
        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]
        min_value_widget_dict = expectation_state["kwargs"].get("min_value")
        min_value_widget = (
            min_value_widget_dict.get("kwarg_widget") if min_value_widget_dict else None
        )
        max_value_widget_dict = expectation_state["kwargs"].get("max_value")
        max_value_widget = (
            max_value_widget_dict.get("kwarg_widget") if max_value_widget_dict else None
        )

        # integer
        if expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["integer"].values())
        ):
            if expectation_type in self.min_max_subtypes["integer"]["positive"]:
                default_min_value = 0
                default_max_value = int(9e300)
            elif expectation_type in self.min_max_subtypes["integer"]["unbounded"]:
                default_min_value = -int(9e300)
                default_max_value = int(9e300)

            if max_value_widget:
                max_value_widget.value = max_value if max_value else default_max_value
            else:
                max_value_widget = widgets.BoundedIntText(
                    value=max_value or default_max_value,
                    max=default_max_value,
                    description="max_value: ",
                    disabled=False,
                )
            if not min_value_widget:
                min_value_widget = widgets.BoundedIntText(
                    min=default_min_value,
                    value=default_min_value,
                    description="min_value: ",
                    disabled=False,
                )
        # float
        elif expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["float"].values())
        ):
            if expectation_type in self.min_max_subtypes["float"]["unit_interval"]:
                default_min_value = 0
                default_max_value = 1
            elif expectation_type in self.min_max_subtypes["float"]["unbounded"]:
                default_min_value = -int(9e300)
                default_max_value = int(9e300)

            if max_value_widget:
                max_value_widget.value = max_value if max_value else default_max_value
            else:
                max_value_widget = widgets.BoundedFloatText(
                    value=max_value or default_max_value,
                    max=default_max_value,
                    step=0.01,
                    description="max_value: ",
                    disabled=False,
                )
            if not min_value_widget:
                min_value_widget = widgets.BoundedFloatText(
                    min=default_min_value,
                    step=0.01,
                    value=default_min_value,
                    description="min_value: ",
                    disabled=False,
                )
        # number or string
        elif expectation_type in list(
            chain.from_iterable(self.min_max_subtypes["mixed"].values())
        ):
            max_value_str = str(max_value) if max_value or max_value == 0 else ""
            if max_value_widget:
                max_value_widget.value = max_value
            else:
                max_value_widget = self.generate_text_widget(
                    value=max_value_str,
                    description="max_value",
                    placeholder="press enter to confirm...",
                )

            @expectation_feedback_widget.capture(clear_output=True)
            def on_max_value_change(change):
                min_max_type = expectation_state["kwargs"]["min_max_type"][
                    "kwarg_widget"
                ].value
                ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                    expectation_state["kwargs"]
                )
                min_value = ge_expectation_kwargs.get("min_value")
                max_value = ge_expectation_kwargs.get("max_value")

                if min_max_type == "string":
                    min_value = str(min_value) if min_value else None
                    max_value = str(max_value) if max_value else None
                elif min_max_type == "number":
                    min_value = float(min_value) if min_value else None
                    max_value = float(max_value) if max_value else None

                if min_value and max_value and min_value > max_value:
                    with expectation_feedback_widget:
                        print(
                            "Error: min_value must be less than or equal to max_value."
                        )
                    return

                ge_expectation_kwargs["min_value"] = min_value
                ge_expectation_kwargs["max_value"] = max_value

                getattr(data_asset, expectation_type)(
                    include_config=True, **ge_expectation_kwargs
                )

            max_value_widget.observe(on_max_value_change, names="value")

        if (
            min_value_widget
            and max_value_widget
            and expectation_type
            not in list(chain.from_iterable(self.min_max_subtypes["mixed"].values()))
        ):

            @expectation_feedback_widget.capture(clear_output=True)
            def on_max_value_change(change):
                ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                    expectation_state["kwargs"]
                )
                getattr(data_asset, expectation_type)(
                    include_config=True, **ge_expectation_kwargs
                )

            max_value_widget.observe(on_max_value_change, names="value")
            min_dl = widgets.link(
                (min_value_widget, "value"), (max_value_widget, "min")
            )
            expectation_state["kwargs"]["min_value"] = {
                "kwarg_widget": min_value_widget
            }

        max_value_widget_dict = (
            {"kwarg_widget": max_value_widget}
            if max_value_widget
            else self.generate_expectation_kwarg_fallback_widget_dict(
                expectation_kwarg_name="max_value", **{"max_value": max_value}
            )
        )
        expectation_state["kwargs"]["max_value"] = max_value_widget_dict

        return max_value_widget_dict

    def generate_value_set_widget_dict(
        self, expectation_state, value_set=None, column=None, **expectation_kwargs
    ):
        if value_set is None:
            value_set = []

        data_asset_name = expectation_state["data_asset_name"]
        data_asset = self.state["data_assets"].get(data_asset_name)["data_asset"]
        expectation_feedback_widget = expectation_state["expectation_feedback_widget"]
        expectation_type = expectation_state["expectation_type"]

        value_set_widget_display = widgets.Box()

        value_set_widget_number_input = self.generate_text_widget(
            value=None,
            description="add number",
            placeholder="press enter to add number...",
            continuous_update=True,
        )
        value_set_widget_string_input = self.generate_text_widget(
            value=None,
            description="add string",
            placeholder="press enter to add string...",
            continuous_update=True,
        )
        value_set_widget_input = widgets.VBox(
            [value_set_widget_string_input, value_set_widget_number_input]
        )
        value_set_label = widgets.HTML(
            value='<div style="text-align:right"><strong>value set: </strong></div>',
            layout={"width": "150px"},
        )
        value_set_widget = widgets.VBox(
            [
                widgets.HBox([value_set_label, value_set_widget_display]),
                value_set_widget_input,
            ]
        )

        widget_dict = {
            "kwarg_widget": value_set_widget,
            "ge_kwarg_value": value_set,
            "widget_display": value_set_widget_display,
            "widget_input": value_set_widget_input,
        }

        value_set_widget_display.children = self.generate_tag_button_list(
            expectation_state=expectation_state,
            tag_list=value_set,
            widget_display=value_set_widget_display,
        )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_value_set_number_input_submit(widget):
            widget_display = expectation_state["kwargs"]["value_set"]["widget_display"]
            value_set = expectation_state["kwargs"]["value_set"]["ge_kwarg_value"]
            new_value = widget.value

            new_value = float(new_value) if "." in new_value else int(new_value)

            if new_value in value_set:
                with expectation_feedback_widget:
                    print("Warning: value already exists in set")
                return

            value_set.append(new_value)
            widget_display.children += (
                self.generate_tag_button(
                    tag=new_value,
                    expectation_state=expectation_state,
                    tag_list=value_set,
                    widget_display=widget_display,
                ),
            )
            widget.value = ""

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        @expectation_feedback_widget.capture(clear_output=True)
        def on_value_set_string_input_submit(widget):
            widget_display = expectation_state["kwargs"]["value_set"]["widget_display"]
            value_set = expectation_state["kwargs"]["value_set"]["ge_kwarg_value"]
            new_value = widget.value

            if new_value in value_set:
                with expectation_feedback_widget:
                    print("Warning: value already exists in set")
                return

            value_set.append(new_value)
            widget_display.children += (
                self.generate_tag_button(
                    tag=new_value,
                    expectation_state=expectation_state,
                    tag_list=value_set,
                    widget_display=widget_display,
                ),
            )
            widget.value = ""

            ge_expectation_kwargs = self.expectation_kwarg_dict_to_ge_kwargs(
                expectation_state["kwargs"]
            )

            getattr(data_asset, expectation_type)(
                include_config=True, **ge_expectation_kwargs
            )

        value_set_widget_number_input.on_submit(on_value_set_number_input_submit)
        value_set_widget_string_input.on_submit(on_value_set_string_input_submit)

        return widget_dict

    def generate_expectation_kwarg_fallback_widget_dict(
        self, expectation_kwarg_name, **expectation_kwargs
    ):
        ge_kwarg_value = expectation_kwargs.get(expectation_kwarg_name)
        warning_message = widgets.HTML(
            value='<div><strong>Warning: </strong>No input widget for kwarg "{expectation_kwarg_name}". To change value, call expectation again with the modified value.</div>'.format(
                expectation_kwarg_name=expectation_kwarg_name
            ),
            layout={"width": "400px", "margin": "0px 0px 0px 10px"},
        )
        static_widget = widgets.Textarea(
            value=str(ge_kwarg_value),
            description="<strong>{expectation_kwarg_name}: </strong>".format(
                expectation_kwarg_name=expectation_kwarg_name
            ),
            style=self.styles.get("description_width"),
            layout={"width": "400px"},
            disabled=True,
        )
        widget_dict = {
            "kwarg_widget": widgets.HBox([static_widget, warning_message]),
            "ge_kwarg_value": ge_kwarg_value,
        }

        return widget_dict

    # widget generators for general info shared between all expectations
    def generate_column_widget(self, column=None):
        return (
            widgets.HTML(
                value="<div><strong>Column: </strong>{column}</div>".format(
                    column=column
                )
            )
            if column
            else None
        )

    def generate_expectation_type_widget(self, expectation_type):
        return widgets.HTML(
            value="<span><strong>Expectation Type: </strong>{expectation_type}</span>".format(
                expectation_type=expectation_type
            )
        )

    def generate_basic_expectation_info_box(
        self,
        data_asset_name,
        expectation_type,
        success_widget,
        validation_time,
        column=None,
    ):
        if column:
            return widgets.VBox(
                [
                    widgets.HTML(
                        value="<div><strong>Data Asset Name: </strong>{data_asset_name}</div>".format(
                            data_asset_name=data_asset_name
                        )
                    ),
                    self.generate_column_widget(column=column),
                    self.generate_expectation_type_widget(expectation_type),
                    success_widget,
                    widgets.HTML(
                        value="<div><strong>Date/Time Validated (UTC): </strong>{validation_time}</div>".format(
                            validation_time=validation_time
                        )
                    ),
                ],
                layout=widgets.Layout(margin="10px", width="40%"),
            )
        else:
            return widgets.VBox(
                [
                    widgets.HTML(
                        value="<div><strong>Data Asset Name: </strong>{data_asset_name}</div>".format(
                            data_asset_name=data_asset_name
                        )
                    ),
                    self.generate_expectation_type_widget(expectation_type),
                    success_widget,
                    widgets.HTML(
                        value="<div><strong>Date/Time Validated (UTC): </strong>{validation_time}</div>".format(
                            validation_time=validation_time
                        )
                    ),
                ],
                layout=widgets.Layout(margin="10px", width="40%"),
            )

    def generate_expectation_result_detail_widgets(self, result=None):
        if result is None:
            result = {}

        result_detail_widgets = []

        for result_title, result_value in result.items():
            result_detail_widgets.append(
                widgets.HTML(
                    value="<span><strong>{}: </strong>{:.2f}</span>".format(
                        result_title, result_value
                    )
                    if type(result_value) is float
                    else "<span><strong>{result_title}: </strong>{result_value}</span>".format(
                        result_title=result_title, result_value=result_value
                    )
                )
            )

        return result_detail_widgets

    # widget generator for complete expectation widget

    def create_expectation_widget(
        self, data_asset, expectation_validation_result, collapsed=False
    ):
        validation_time = datetime.datetime.now(datetime.timezone.utc).strftime(
            "%Y-%m-%d %H:%M"
        )
        data_asset_name = data_asset.data_asset_name
        data_asset_state = self.state["data_assets"].get(data_asset_name)
        expectation_type = (
            expectation_validation_result.expectation_config.expectation_type
        )
        expectation_kwargs = expectation_validation_result.expectation_config["kwargs"]
        column = expectation_kwargs.get("column")

        if data_asset_state:
            existing_expectation_state = self.get_expectation_state(
                data_asset_name=data_asset_name,
                expectation_type=expectation_type,
                column=column,
            )
            if existing_expectation_state:
                return self.update_expectation_state(
                    existing_expectation_state,
                    expectation_validation_result,
                    validation_time,
                )
        else:
            self.initialize_data_asset_state(data_asset)

        # success_widget
        success = expectation_validation_result.success
        success_widget = widgets.HTML(
            value="<span><strong>Success: </strong>{success}</span>".format(
                success=str(success)
            )
        )

        # widget with result details
        result_detail_widget = widgets.VBox(
            children=self.generate_expectation_result_detail_widgets(
                result=expectation_validation_result.result
            )
        )

        # accordion container for result_detail_widget
        result_detail_accordion = widgets.Accordion(children=[result_detail_widget])
        result_detail_accordion.set_title(0, "Validation Result Details")
        result_detail_accordion.selected_index = None

        # widget for displaying kwarg widget exceptions
        expectation_feedback_widget = widgets.Output()
        exception_accordion = widgets.Accordion(children=[expectation_feedback_widget])
        exception_accordion.set_title(0, "Exceptions/Warnings")

        # create scaffold dict for expectation state
        expectation_state = {
            "success": success_widget,
            "expectation_type": expectation_type,
            "expectation_feedback_widget": expectation_feedback_widget,
            "data_asset_name": data_asset_name,
            "result_detail_widget": result_detail_widget,
            "kwargs": {},
        }

        if column:
            expectation_state["kwargs"]["column"] = column

        remove_expectation_button = self.generate_remove_expectation_button(
            expectation_state=expectation_state
        )

        # widget with basic info (column, expectation_type)
        basic_expectation_info_box = self.generate_basic_expectation_info_box(
            data_asset_name, expectation_type, success_widget, validation_time, column
        )

        # collect widget dicts for kwarg inputs
        expectation_kwarg_input_widget_dicts = []
        for expectation_kwarg_name in self.expectation_kwarg_field_names.get(
            expectation_type
        ):
            widget_dict_generator = getattr(
                self,
                "generate_{expectation_kwarg_name}_widget_dict".format(
                    expectation_kwarg_name=expectation_kwarg_name
                ),
                None,
            )
            widget_dict = (
                widget_dict_generator(
                    expectation_state=expectation_state, **expectation_kwargs
                )
                if widget_dict_generator
                else self.generate_expectation_kwarg_fallback_widget_dict(
                    expectation_kwarg_name=expectation_kwarg_name, **expectation_kwargs
                )
            )
            expectation_state["kwargs"][expectation_kwarg_name] = widget_dict
            expectation_kwarg_input_widget_dicts.append(widget_dict)

        # container for kwarg input widgets
        kwarg_widgets = [
            dict["kwarg_widget"] for dict in expectation_kwarg_input_widget_dicts
        ]
        expectation_kwarg_input_box = widgets.VBox(
            children=kwarg_widgets, layout=widgets.Layout(margin="10px", width="60%")
        )

        # top-level widget container
        expectation_editor_widget_content = widgets.VBox(
            [
                widgets.HBox([basic_expectation_info_box, expectation_kwarg_input_box]),
                exception_accordion,
                result_detail_accordion,
                remove_expectation_button,
            ]
        )
        expectation_editor_widget = widgets.Accordion(
            children=[expectation_editor_widget_content],
            layout=widgets.Layout(
                border="2px solid {color}".format(color="green" if success else "red"),
                margin="5px",
            ),
        )
        expectation_editor_title = (
            "{column} | {expectation_type}".format(
                column=column, expectation_type=expectation_type
            )
            if column
            else "non_column_expectation | {expectation_type}".format(
                expectation_type=expectation_type
            )
        )
        expectation_editor_widget.set_title(0, expectation_editor_title)

        if collapsed:
            expectation_editor_widget.selected_index = None

        # complete expectation state scaffold
        expectation_state["editor_widget"] = expectation_editor_widget

        # set expectation state
        self.set_expectation_state(data_asset, expectation_state, column)

        # update expectation_suite_editor
        expectation_suite_editor = self.state["data_assets"][data_asset_name].get(
            "expectation_suite_editor"
        )
        if expectation_suite_editor:
            expectation_suite = data_asset.get_expectation_suite(
                discard_failed_expectations=False
            )
            expectation_suite_editor.children = (
                self.generate_expectation_suite_editor_widgets(
                    data_asset, expectation_suite
                )
            )

        return expectation_editor_widget

    # expectation_suite_editing
    def get_column_names(self, data_asset_name):
        data_asset_state = self.state["data_assets"].get(data_asset_name)
        if not data_asset_state:
            return []
        data_asset_expectations = data_asset_state.get("expectations")
        if not data_asset_expectations:
            return []

        column_names = [
            column_name
            for column_name in data_asset_expectations.keys()
            if column_name != "non_column_expectations"
        ]
        column_names.sort()
        if "non_column_expectations" in data_asset_expectations.keys():
            column_names.insert(0, "non_column_expectations")

        return column_names

    def get_expectation_types(self, data_asset_name):
        data_asset_state = self.state["data_assets"].get(data_asset_name)
        data_asset_expectations = data_asset_state.get("expectations")
        if not data_asset_state or not data_asset_expectations:
            return []

        expectation_types = []

        for expectation_group in data_asset_expectations.values():
            expectation_types += list(expectation_group.keys())

        return list(set(expectation_types))

    def generate_expectation_suite_editor_widgets(self, data_asset, expectation_suite):
        data_asset_name = data_asset.data_asset_name
        column_names = self.get_column_names(data_asset_name)
        column_accordions = []
        data_asset_state = self.state["data_assets"].get(data_asset_name, {})
        data_asset_expectations = data_asset_state.get("expectations", {})

        # TODO: Deprecate "great_expectations.__version__"
        ge_version = expectation_suite.get("meta").get(
            "great_expectations_version"
        ) or expectation_suite.get("meta").get("great_expectations.__version__")
        column_count = len(column_names)
        expectation_count = len(expectation_suite["expectations"])

        summary_widget_content = """\
            <div>
                <h1>Expectation Suite Editor</h1>
                <hr>
                <ul>
                    <li><strong>Data Asset Name</strong>: {data_asset_name}</li>
                    <li><strong>Great Expectations Version</strong>: {ge_version}</li>
                    <li><strong>Column Count</strong>: {column_count}</li>
                    <li><strong>Expectation Count</strong>: {expectation_count}</li>
                </ul>
                <hr>
                <h2>Columns:</h2>
            </div>\
        """.format(
            data_asset_name=data_asset_name,
            ge_version=ge_version,
            column_count=column_count,
            expectation_count=expectation_count,
        )
        summary_widget = widgets.HTML(
            value=summary_widget_content,
            layout=widgets.Layout(
                margin="10px",
            ),
        )

        for column_name in column_names:
            expectation_editor_widgets = []
            column_expectations = data_asset_expectations[column_name]
            expectation_types = list(column_expectations.keys())
            expectation_types.sort()
            for expectation_type in expectation_types:
                editor_widget = column_expectations[expectation_type]["editor_widget"]
                editor_widget.selected_index = None
                expectation_editor_widgets.append(editor_widget)
            column_accordion_content = widgets.VBox(children=expectation_editor_widgets)
            column_accordion = widgets.Accordion(children=[column_accordion_content])
            column_accordion.set_title(0, column_name)
            column_accordion.selected_index = None
            column_accordions.append(column_accordion)

        return [summary_widget] + column_accordions

    def edit_expectation_suite(self, data_asset):
        data_asset_name = data_asset.data_asset_name
        expectation_suite = data_asset.get_expectation_suite(
            discard_failed_expectations=False
        )
        expectations = expectation_suite.get("expectations")

        ################### editor widgets
        for expectation in expectations:
            expectation_type = expectation.get("expectation_type")
            expectation_kwargs = expectation.get("kwargs")
            editor_widget = getattr(data_asset, expectation_type)(
                include_config=True, **expectation_kwargs
            )

        expectation_suite_editor_widgets = (
            self.generate_expectation_suite_editor_widgets(
                data_asset, expectation_suite
            )
        )
        ###################
        expectation_suite_editor = widgets.VBox(
            children=expectation_suite_editor_widgets,
            layout=widgets.Layout(
                border="2px solid black",
            ),
        )

        self.state["data_assets"][data_asset_name][
            "expectation_suite_editor"
        ] = expectation_suite_editor

        return expectation_suite_editor



import logging

from .expectation_suite import (
    ExpectationConfiguration,
    ExpectationConfigurationSchema,
    ExpectationSuite,
    ExpectationSuiteSchema,
    expectationSuiteSchema,
)
from .expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationSuiteValidationResultSchema,
    ExpectationValidationResult,
    ExpectationValidationResultSchema,
    expectationSuiteValidationResultSchema,
    expectationValidationResultSchema,
    get_metric_kwargs_id,
)
from .id_dict import IDDict
from .run_identifier import RunIdentifier, RunIdentifierSchema
from .urn import ge_urn

__all__ = [
    ExpectationConfiguration,
    ExpectationConfigurationSchema,
    ExpectationSuite,
    ExpectationSuiteSchema,
    expectationSuiteSchema,
    ExpectationSuiteValidationResult,
    ExpectationSuiteValidationResultSchema,
    ExpectationValidationResult,
    ExpectationValidationResultSchema,
    expectationSuiteValidationResultSchema,
    expectationValidationResultSchema,
    get_metric_kwargs_id,
    IDDict,
    RunIdentifier,
    RunIdentifierSchema,
    ge_urn,
]

logger = logging.getLogger(__name__)

RESULT_FORMATS = ["BOOLEAN_ONLY", "BASIC", "COMPLETE", "SUMMARY"]



from .configured_asset_file_path_data_connector import (
    ConfiguredAssetFilePathDataConnector,
)
from .configured_asset_filesystem_data_connector import (
    ConfiguredAssetFilesystemDataConnector,
)
from .configured_asset_s3_data_connector import ConfiguredAssetS3DataConnector
from .configured_asset_sql_data_connector import ConfiguredAssetSqlDataConnector
from .data_connector import DataConnector
from .file_path_data_connector import FilePathDataConnector
from .inferred_asset_file_path_data_connector import InferredAssetFilePathDataConnector
from .inferred_asset_filesystem_data_connector import (
    InferredAssetFilesystemDataConnector,
)
from .inferred_asset_s3_data_connector import InferredAssetS3DataConnector
from .inferred_asset_sql_data_connector import InferredAssetSqlDataConnector
from .runtime_data_connector import RuntimeDataConnector



from .column_distinct_values import ColumnDistinctValues, ColumnDistinctValuesCount
from .column_histogram import ColumnHistogram
from .column_max import ColumnMax
from .column_mean import ColumnMean
from .column_median import ColumnMedian
from .column_min import ColumnMin
from .column_most_common_value import ColumnMostCommonValue
from .column_parameterized_distribution_ks_test_p_value import (
    ColumnParameterizedDistributionKSTestPValue,
)
from .column_partition import ColumnPartition
from .column_proportion_of_unique_values import ColumnUniqueProportion
from .column_quantile_values import ColumnQuantileValues
from .column_standard_deviation import ColumnStandardDeviation
from .column_sum import ColumnSum
from .column_value_counts import ColumnValueCounts
from .column_values_between_count import ColumnValuesBetweenCount



import logging
from typing import Iterator, List, Optional

import great_expectations.exceptions as ge_exceptions
from great_expectations.core.batch import BatchDefinition, BatchRequest
from great_expectations.datasource.data_connector.data_connector import DataConnector
from great_expectations.datasource.data_connector.partition_query import (
    PartitionQuery,
    build_partition_query,
)
from great_expectations.datasource.data_connector.sorter import Sorter
from great_expectations.datasource.data_connector.util import (
    batch_definition_matches_batch_request,
    build_sorters_from_config,
    map_batch_definition_to_data_reference_string_using_regex,
    map_data_reference_string_to_batch_definition_list_using_regex,
)
from great_expectations.datasource.types import PathBatchSpec
from great_expectations.execution_engine import ExecutionEngine

logger = logging.getLogger(__name__)


class FilePathDataConnector(DataConnector):
    """
    Base-class for DataConnector that are designed for connecting to filesystem-like data, which can include
    files on disk, but also S3 and GCS.

    *Note*: FilePathDataConnector is not meant to be used on its own, but extended. Currently
    ConfiguredAssetFilePathDataConnector and InferredAssetFilePathDataConnector are subclasses of
    FilePathDataConnector.
    """

    def __init__(
        self,
        name: str,
        datasource_name: str,
        execution_engine: Optional[ExecutionEngine] = None,
        default_regex: Optional[dict] = None,
        sorters: Optional[list] = None,
    ):
        """
        Base class for DataConnectors that connect to filesystem-like data. This class supports the configuration of default_regex
        and sorters for filtering and sorting data_references.

        Args:
            name (str): name of FilePathDataConnector
            datasource_name (str): Name of datasource that this DataConnector is connected to
            execution_engine (ExecutionEngine): Execution Engine object to actually read the data
            default_regex (dict): Optional dict the filter and organize the data_references.
            sorters (list): Optional list if you want to sort the data_references
        """
        logger.debug(f'Constructing FilePathDataConnector "{name}".')

        super().__init__(
            name=name,
            datasource_name=datasource_name,
            execution_engine=execution_engine,
        )

        if default_regex is None:
            default_regex = {}
        self._default_regex = default_regex

        self._sorters = build_sorters_from_config(config_list=sorters)
        self._validate_sorters_configuration()

    @property
    def sorters(self) -> Optional[dict]:
        return self._sorters

    def _get_data_reference_list_from_cache_by_data_asset_name(
        self, data_asset_name: str
    ) -> List[str]:
        """
        Fetch data_references corresponding to data_asset_name from the cache.
        """
        regex_config: dict = self._get_regex_config(data_asset_name=data_asset_name)
        pattern: str = regex_config["pattern"]
        group_names: List[str] = regex_config["group_names"]

        batch_definition_list = self.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name=self.datasource_name,
                data_connector_name=self.name,
                data_asset_name=data_asset_name,
            )
        )

        path_list: List[str] = [
            map_batch_definition_to_data_reference_string_using_regex(
                batch_definition=batch_definition,
                regex_pattern=pattern,
                group_names=group_names,
            )
            for batch_definition in batch_definition_list
        ]

        # TODO: Sort with a real sorter here
        path_list.sort()

        return path_list

    def get_batch_definition_list_from_batch_request(
        self,
        batch_request: BatchRequest,
    ) -> List[BatchDefinition]:
        """
        Retrieve batch_definitions and that match batch_request.

        First retrieves all batch_definitions that match batch_request
            - if batch_request also has a partition_query, then select batch_definitions that match partition_query.
            - if data_connector has sorters configured, then sort the batch_definition list before returning.

        Args:
            batch_request (BatchRequest): BatchRequest to process

        Returns:
            A list of BatchDefinition objects that match BatchRequest

        """
        self._validate_batch_request(batch_request=batch_request)

        if self._data_references_cache is None:
            self._refresh_data_references_cache()

        batch_definition_list: List[BatchDefinition] = list(
            filter(
                lambda batch_definition: batch_definition_matches_batch_request(
                    batch_definition=batch_definition, batch_request=batch_request
                ),
                self._get_batch_definition_list_from_cache(),
            )
        )

        if batch_request.partition_request is not None:
            partition_query_obj: PartitionQuery = build_partition_query(
                partition_request_dict=batch_request.partition_request
            )
            batch_definition_list = partition_query_obj.select_from_partition_request(
                batch_definition_list=batch_definition_list
            )

        if len(self.sorters) > 0:
            sorted_batch_definition_list = self._sort_batch_definition_list(
                batch_definition_list=batch_definition_list
            )
            return sorted_batch_definition_list
        else:
            return batch_definition_list

    def _sort_batch_definition_list(
        self, batch_definition_list: List[BatchDefinition]
    ) -> List[BatchDefinition]:
        """
        Use configured sorters to sort batch_definition

        Args:
            batch_definition_list (list): list of batch_definitions to sort

        Returns:
            sorted list of batch_definitions

        """
        sorters: Iterator[Sorter] = reversed(list(self.sorters.values()))
        for sorter in sorters:
            batch_definition_list = sorter.get_sorted_batch_definitions(
                batch_definitions=batch_definition_list
            )
        return batch_definition_list

    def _map_data_reference_to_batch_definition_list(
        self, data_reference: str, data_asset_name: str = None
    ) -> Optional[List[BatchDefinition]]:
        regex_config: dict = self._get_regex_config(data_asset_name=data_asset_name)
        pattern: str = regex_config["pattern"]
        group_names: List[str] = regex_config["group_names"]
        return map_data_reference_string_to_batch_definition_list_using_regex(
            datasource_name=self.datasource_name,
            data_connector_name=self.name,
            data_asset_name=data_asset_name,
            data_reference=data_reference,
            regex_pattern=pattern,
            group_names=group_names,
        )

    def _map_batch_definition_to_data_reference(
        self, batch_definition: BatchDefinition
    ) -> str:
        data_asset_name: str = batch_definition.data_asset_name
        regex_config: dict = self._get_regex_config(data_asset_name=data_asset_name)
        pattern: str = regex_config["pattern"]
        group_names: List[str] = regex_config["group_names"]
        return map_batch_definition_to_data_reference_string_using_regex(
            batch_definition=batch_definition,
            regex_pattern=pattern,
            group_names=group_names,
        )

    def build_batch_spec(self, batch_definition: BatchDefinition) -> PathBatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (BatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec = super().build_batch_spec(batch_definition=batch_definition)
        return PathBatchSpec(batch_spec)

    def _generate_batch_spec_parameters_from_batch_definition(
        self, batch_definition: BatchDefinition
    ) -> dict:
        path: str = self._map_batch_definition_to_data_reference(
            batch_definition=batch_definition
        )
        if not path:
            raise ValueError(
                f"""No data reference for data asset name "{batch_definition.data_asset_name}" matches the given
partition definition {batch_definition.partition_definition} from batch definition {batch_definition}.
                """
            )
        path = self._get_full_file_path(
            path=path, data_asset_name=batch_definition.data_asset_name
        )
        return {"path": path}

    def _validate_batch_request(self, batch_request: BatchRequest):
        super()._validate_batch_request(batch_request)
        self._validate_sorters_configuration(
            data_asset_name=batch_request.data_asset_name
        )

    def _validate_sorters_configuration(self, data_asset_name: Optional[str] = None):
        if self.sorters is not None and len(self.sorters) > 0:
            # data_asset_name: str = batch_request.data_asset_name
            regex_config: dict = self._get_regex_config(data_asset_name=data_asset_name)
            group_names: List[str] = regex_config["group_names"]
            if any(
                [sorter_name not in group_names for sorter_name in self.sorters.keys()]
            ):
                raise ge_exceptions.DataConnectorError(
                    f"""DataConnector "{self.name}" specifies one or more sort keys that do not appear among the
configured group_name.
                    """
                )
            if len(group_names) < len(self.sorters):
                raise ge_exceptions.DataConnectorError(
                    f"""DataConnector "{self.name}" is configured with {len(group_names)} group names;
this is fewer than number of sorters specified, which is {len(self.sorters)}.
                    """
                )

    def _get_batch_definition_list_from_cache(self) -> List[BatchDefinition]:
        raise NotImplementedError

    def _get_regex_config(self, data_asset_name: Optional[str] = None) -> dict:
        raise NotImplementedError

    def _get_full_file_path(
        self, path: str, data_asset_name: Optional[str] = None
    ) -> str:
        raise NotImplementedError



from typing import Any, Dict, Optional, Tuple

import numpy as np

from great_expectations.core import ExpectationConfiguration
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
)
from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric import (
    ColumnMetricProvider,
    column_aggregate_partial,
    column_aggregate_value,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.validator.validation_graph import MetricConfiguration


class ColumnPartition(ColumnMetricProvider):
    metric_name = "column.partition"
    value_keys = ("bins", "n_bins", "allow_relative_error")
    default_kwarg_values = {
        "bins": "uniform",
        "n_bins": 10,
        "allow_relative_error": False,
    }

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):
        bins = metric_value_kwargs.get("bins", cls.default_kwarg_values["bins"])
        n_bins = metric_value_kwargs.get("n_bins", cls.default_kwarg_values["n_bins"])
        return _get_column_partition_using_metrics(bins, n_bins, metrics)

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):
        bins = metric_value_kwargs.get("bins", cls.default_kwarg_values["bins"])
        n_bins = metric_value_kwargs.get("n_bins", cls.default_kwarg_values["n_bins"])
        return _get_column_partition_using_metrics(bins, n_bins, metrics)

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):
        bins = metric_value_kwargs.get("bins", cls.default_kwarg_values["bins"])
        n_bins = metric_value_kwargs.get("n_bins", cls.default_kwarg_values["n_bins"])
        return _get_column_partition_using_metrics(bins, n_bins, metrics)

    @classmethod
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        bins = metric.metric_value_kwargs.get("bins", cls.default_kwarg_values["bins"])
        n_bins = metric.metric_value_kwargs.get(
            "n_bins", cls.default_kwarg_values["n_bins"]
        )
        allow_relative_error = metric.metric_value_kwargs["allow_relative_error"]

        if bins == "uniform":
            return {
                "column.min": MetricConfiguration(
                    "column.min", metric.metric_domain_kwargs
                ),
                "column.max": MetricConfiguration(
                    "column.max", metric.metric_domain_kwargs
                ),
            }
        elif bins in ["ntile", "quantile", "percentile"]:
            return {
                "column.quantile_values": MetricConfiguration(
                    "column.quantile_values",
                    metric.metric_domain_kwargs,
                    {
                        "quantiles": np.linspace(
                            start=0, stop=1, num=n_bins + 1
                        ).tolist(),
                        "allow_relative_error": allow_relative_error,
                    },
                )
            }
        elif bins == "auto":
            return {
                "column_values.nonnull.count": MetricConfiguration(
                    "column_values.nonnull.count",
                    metric.metric_domain_kwargs,
                ),
                "column.quantile_values": MetricConfiguration(
                    "column.quantile_values",
                    metric.metric_domain_kwargs,
                    {
                        "quantiles": (0.0, 0.25, 0.75, 1.0),
                        "allow_relative_error": allow_relative_error,
                    },
                ),
            }
        else:
            raise ValueError("Invalid parameter for bins argument")


def _get_column_partition_using_metrics(bins, n_bins, _metrics):
    if bins == "uniform":
        min_ = _metrics["column.min"]
        max_ = _metrics["column.max"]
        # PRECISION NOTE: some implementations of quantiles could produce
        # varying levels of precision (e.g. a NUMERIC column producing
        # Decimal from a SQLAlchemy source, so we cast to float for numpy)
        bins = np.linspace(start=float(min_), stop=float(max_), num=n_bins + 1).tolist()
    elif bins in ["ntile", "quantile", "percentile"]:
        bins = _metrics["column.quantile_values"]
    elif bins == "auto":
        # Use the method from numpy histogram_bin_edges
        nonnull_count = _metrics["column_values.nonnull.count"]
        sturges = np.log2(nonnull_count + 1)
        min_, _25, _75, max_ = _metrics["column.quantile_values"]
        iqr = _75 - _25
        if iqr < 1e-10:  # Consider IQR 0 and do not use variance-based estimator
            n_bins = sturges
        else:
            fd = (2 * float(iqr)) / (nonnull_count ** (1 / 3))
            n_bins = max(int(np.ceil(sturges)), int(np.ceil(float(max_ - min_) / fd)))
        bins = np.linspace(start=float(min_), stop=float(max_), num=n_bins + 1).tolist()
    else:
        raise ValueError("Invalid parameter for bins argument")
    return bins



import logging

from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
)
from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.expectations.metrics.util import get_dialect_regex_expression

logger = logging.getLogger(__name__)


class ColumnValuesNotMatchRegex(ColumnMapMetricProvider):
    condition_metric_name = "column_values.not_match_regex"
    condition_value_keys = ("regex",)

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column, regex, **kwargs):
        return ~column.astype(str).str.contains(regex)

    @column_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column, regex, _dialect, **kwargs):
        regex_expression = get_dialect_regex_expression(
            column, regex, _dialect, positive=False
        )
        if regex_expression is None:
            logger.warning("Regex is not supported for dialect %s" % str(_dialect))
            raise NotImplementedError

        return regex_expression

    @column_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column, regex, **kwargs):
        return ~column.rlike(regex)



from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from dateutil.parser import parse

from great_expectations.core.batch import Batch
from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
)
from great_expectations.expectations.util import render_evaluation_parameter_string

from ...render.renderer.renderer import renderer
from ...render.types import RenderedStringTemplateContent
from ...render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from ..expectation import (
    ColumnMapExpectation,
    Expectation,
    InvalidExpectationConfigurationError,
    TableExpectation,
    _format_map_output,
)
from ..registry import extract_metrics, get_metric_kwargs

try:
    import sqlalchemy as sa
except ImportError:
    pass


class ExpectColumnPairValuesAToBeGreaterThanB(TableExpectation):
    """
    Expect values in column A to be greater than column B.

    Args:
        column_A (str): The first column name
        column_B (str): The second column name
        or_equal (boolean or None): If True, then values can be equal, not strictly greater

    Keyword Args:
        allow_cross_type_comparisons (boolean or None) : If True, allow comparisons between types (e.g. integer and\
            string). Otherwise, attempting such comparisons will raise an exception.

    Keyword Args:
        ignore_row_if (str): "both_values_are_missing", "either_value_is_missing", "neither

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: `BOOLEAN_ONLY`, `BASIC`, `COMPLETE`, or `SUMMARY`.
            For more detail, see :ref:`result_format <result_format>`.
        include_config (boolean): \
            If True, then include the expectation config as part of the result object. \
            For more detail, see :ref:`include_config`.
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see :ref:`catch_exceptions`.
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see :ref:`meta`.

    Returns:
        An ExpectationSuiteValidationResult

        Exact fields vary depending on the values passed to :ref:`result_format <result_format>` and
        :ref:`include_config`, :ref:`catch_exceptions`, and :ref:`meta`.

    """

    metric_dependencies = ("column_a_greater_than_b",)
    success_keys = (
        "column_A",
        "column_B",
        "ignore_row_if",
        "parse_strings_as_datetimes",
        "allow_cross_type_comparisons",
        "or_equal",
    )

    default_kwarg_values = {
        "column_A": None,
        "column_B": None,
        "or_equal": None,
        "parse_strings_as_datetimes": False,
        "allow_cross_type_comparisons": None,
        "ignore_row_if": "both_values_are_missing",
        "row_condition": None,
        "condition_parser": None,  # we expect this to be explicitly set whenever a row_condition is passed
        "result_format": "BASIC",
        "include_config": True,
        "catch_exceptions": True,
    }

    def validate_configuration(self, configuration: Optional[ExpectationConfiguration]):
        super().validate_configuration(configuration)
        if configuration is None:
            configuration = self.configuration
        try:
            assert (
                "column_A" in configuration.kwargs
                and "column_B" in configuration.kwargs
            ), "both columns must be provided"
        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))
        return True

    # @PandasExecutionEngine.metric(
    #     metric_name="column_a_greater_than_b",
    #     metric_domain_keys=ColumnMapExpectation.domain_keys,
    #     metric_value_keys=("column_A", "column_B"),
    #     metric_dependencies=tuple(),
    # )
    def _pandas_column_a_greater_than_b(
        self,
        batches: Dict[str, Batch],
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict,
        runtime_configuration: dict = None,
    ):
        """Metric which returns all columns in a dataframe"""
        df = execution_engine.get_domain_dataframe(
            domain_kwargs=metric_domain_kwargs, batches=batches
        )
        # Initialization of necessary value kwargs
        allow_cross_type_comparisons = None
        parse_strings_as_datetimes = None
        or_equal = None

        column_A = df[metric_value_kwargs["column_A"]]
        column_B = df[metric_value_kwargs["column_B"]]

        # If value kwargs are given that could impact outcome, initializing them
        if allow_cross_type_comparisons in metric_value_kwargs:
            allow_cross_type_comparisons = metric_value_kwargs[
                "allow_cross_type_comparisons"
            ]

        if parse_strings_as_datetimes in metric_value_kwargs:
            parse_strings_as_datetimes = metric_value_kwargs[
                "parse_strings_as_datetimes"
            ]

        if or_equal in metric_value_kwargs:
            or_equal = metric_value_kwargs["or_equal"]

        if allow_cross_type_comparisons:
            column_A = column_A.apply(str)
            column_B = column_B.apply(str)

        if parse_strings_as_datetimes:
            temp_column_A = column_A.map(parse)
            temp_column_B = column_B.map(parse)

        else:
            temp_column_A = column_A
            temp_column_B = column_B

        if or_equal:
            return temp_column_A >= temp_column_B
        else:
            return temp_column_A > temp_column_B

    @classmethod
    @renderer(renderer_type="renderer.prescriptive")
    @render_evaluation_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration=None,
        result=None,
        language=None,
        runtime_configuration=None,
        **kwargs
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name", True)
        include_column_name = (
            include_column_name if include_column_name is not None else True
        )
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            [
                "column_A",
                "column_B",
                "parse_strings_as_datetimes",
                "ignore_row_if",
                "mostly",
                "or_equal",
                "row_condition",
                "condition_parser",
            ],
        )

        if (params["column_A"] is None) or (params["column_B"] is None):
            template_str = "$column has a bogus `expect_column_pair_values_A_to_be_greater_than_B` expectation."
            params["row_condition"] = None

        if params["mostly"] is None:
            if params["or_equal"] in [None, False]:
                template_str = "Values in $column_A must always be greater than those in $column_B."
            else:
                template_str = "Values in $column_A must always be greater than or equal to those in $column_B."
        else:
            params["mostly_pct"] = num_to_str(
                params["mostly"] * 100, precision=15, no_scientific=True
            )
            # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")
            if params["or_equal"] in [None, False]:
                template_str = "Values in $column_A must be greater than those in $column_B, at least $mostly_pct % of the time."
            else:
                template_str = "Values in $column_A must be greater than or equal to those in $column_B, at least $mostly_pct % of the time."

        if params.get("parse_strings_as_datetimes"):
            template_str += " Values should be parsed as datetimes."

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = (
                conditional_template_str
                + ", then "
                + template_str[0].lower()
                + template_str[1:]
            )
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    # @Expectation.validates(metric_dependencies=metric_dependencies)
    def _validates(
        self,
        configuration: ExpectationConfiguration,
        metrics: Dict,
        runtime_configuration: dict = None,
        execution_engine: ExecutionEngine = None,
    ):
        metric_dependencies = self.get_validation_dependencies(
            configuration, execution_engine, runtime_configuration
        )["metrics"]
        metric_vals = extract_metrics(
            metric_dependencies, metrics, configuration, runtime_configuration
        )
        equal_columns = metric_vals["column_a_greater_than_b"]

        if runtime_configuration:
            result_format = runtime_configuration.get(
                "result_format",
                configuration.kwargs.get(
                    "result_format", self.default_kwarg_values.get("result_format")
                ),
            )
        else:
            result_format = configuration.kwargs.get(
                "result_format", self.default_kwarg_values.get("result_format")
            )

        return {"success": equal_columns.all()}



import atexit
import datetime
import json
import logging
import platform
import signal
import sys
import threading
from functools import wraps
from queue import Queue

import jsonschema
import requests

from great_expectations import __version__ as ge_version
from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer
from great_expectations.core.usage_statistics.anonymizers.batch_anonymizer import (
    BatchAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.data_docs_site_anonymizer import (
    DataDocsSiteAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.datasource_anonymizer import (
    DatasourceAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.execution_engine_anonymizer import (
    ExecutionEngineAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.expectation_suite_anonymizer import (
    ExpectationSuiteAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.store_anonymizer import (
    StoreAnonymizer,
)
from great_expectations.core.usage_statistics.anonymizers.validation_operator_anonymizer import (
    ValidationOperatorAnonymizer,
)
from great_expectations.core.usage_statistics.schemas import (
    usage_statistics_record_schema,
)
from great_expectations.core.util import nested_update

STOP_SIGNAL = object()

logger = logging.getLogger(__name__)

_anonymizers = dict()


class UsageStatisticsHandler:
    def __init__(self, data_context, data_context_id, usage_statistics_url):
        self._url = usage_statistics_url

        self._data_context_id = data_context_id
        self._data_context_instance_id = data_context.instance_id
        self._data_context = data_context
        self._ge_version = ge_version

        self._message_queue = Queue()
        self._worker = threading.Thread(target=self._requests_worker, daemon=True)
        self._worker.start()
        self._datasource_anonymizer = DatasourceAnonymizer(data_context_id)
        self._execution_engine_anonymizer = ExecutionEngineAnonymizer(data_context_id)
        self._store_anonymizer = StoreAnonymizer(data_context_id)
        self._validation_operator_anonymizer = ValidationOperatorAnonymizer(
            data_context_id
        )
        self._data_docs_sites_anonymizer = DataDocsSiteAnonymizer(data_context_id)
        self._batch_anonymizer = BatchAnonymizer(data_context_id)
        self._expectation_suite_anonymizer = ExpectationSuiteAnonymizer(data_context_id)
        try:
            self._sigterm_handler = signal.signal(signal.SIGTERM, self._teardown)
        except ValueError:
            # if we are not the main thread, we don't get to ask for signal handling.
            self._sigterm_handler = None
        try:
            self._sigint_handler = signal.signal(signal.SIGINT, self._teardown)
        except ValueError:
            # if we are not the main thread, we don't get to ask for signal handling.
            self._sigint_handler = None

        atexit.register(self._close_worker)

    def _teardown(self, signum: int, frame):
        self._close_worker()
        if signum == signal.SIGTERM and self._sigterm_handler:
            self._sigterm_handler(signum, frame)
        if signum == signal.SIGINT and self._sigint_handler:
            self._sigint_handler(signum, frame)

    def _close_worker(self):
        self._message_queue.put(STOP_SIGNAL)
        self._worker.join()

    def _requests_worker(self):
        session = requests.Session()
        while True:
            message = self._message_queue.get()
            if message == STOP_SIGNAL:
                self._message_queue.task_done()
                return
            try:
                res = session.post(self._url, json=message, timeout=2)
                logger.debug(
                    "Posted usage stats: message status " + str(res.status_code)
                )
                if res.status_code != 201:
                    logger.debug(
                        "Server rejected message: ", json.dumps(message, indent=2)
                    )
            except requests.exceptions.Timeout:
                logger.debug("Timeout while sending usage stats message.")
            except Exception as e:
                logger.debug("Unexpected error posting message: " + str(e))
            finally:
                self._message_queue.task_done()

    def send_usage_message(self, event, event_payload=None, success=None):
        """send a usage statistics message."""
        try:
            message = {
                "event": event,
                "event_payload": event_payload or {},
                "success": success,
            }

            self.emit(message)
        except Exception:
            pass

    def build_init_payload(self):
        """Adds information that may be available only after full data context construction, but is useful to
        calculate only one time (for example, anonymization)."""
        expectation_suites = [
            self._data_context.get_expectation_suite(expectation_suite_name)
            for expectation_suite_name in self._data_context.list_expectation_suite_names()
        ]
        return {
            "platform.system": platform.system(),
            "platform.release": platform.release(),
            "version_info": str(sys.version_info),
            "anonymized_datasources": [
                self._datasource_anonymizer.anonymize_datasource_info(
                    datasource_name, datasource_config
                )
                for datasource_name, datasource_config in self._data_context._project_config_with_variables_substituted.datasources.items()
            ],
            "anonymized_stores": [
                self._store_anonymizer.anonymize_store_info(store_name, store_obj)
                for store_name, store_obj in self._data_context.stores.items()
            ],
            "anonymized_validation_operators": [
                self._validation_operator_anonymizer.anonymize_validation_operator_info(
                    validation_operator_name=validation_operator_name,
                    validation_operator_obj=validation_operator_obj,
                )
                for validation_operator_name, validation_operator_obj in self._data_context.validation_operators.items()
            ],
            "anonymized_data_docs_sites": [
                self._data_docs_sites_anonymizer.anonymize_data_docs_site_info(
                    site_name=site_name, site_config=site_config
                )
                for site_name, site_config in self._data_context._project_config_with_variables_substituted.data_docs_sites.items()
            ],
            "anonymized_expectation_suites": [
                self._expectation_suite_anonymizer.anonymize_expectation_suite_info(
                    expectation_suite
                )
                for expectation_suite in expectation_suites
            ],
        }

    def build_envelope(self, message):
        message["version"] = "1.0.0"
        message["event_time"] = (
            datetime.datetime.now(datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%S.%f"
            )[:-3]
            + "Z"
        )
        message["data_context_id"] = self._data_context_id
        message["data_context_instance_id"] = self._data_context_instance_id
        message["ge_version"] = self._ge_version
        return message

    def validate_message(self, message, schema):
        try:
            jsonschema.validate(message, schema=schema)
            return True
        except jsonschema.ValidationError as e:
            logger.debug("invalid message: " + str(e))
            return False

    def emit(self, message):
        """
        Emit a message.
        """
        try:
            if message["event"] == "data_context.__init__":
                message["event_payload"] = self.build_init_payload()
            message = self.build_envelope(message)
            if not self.validate_message(
                message, schema=usage_statistics_record_schema
            ):
                return
            self._message_queue.put(message)
        # noinspection PyBroadException
        except Exception as e:
            # We *always* tolerate *any* error in usage statistics
            logger.debug(e)


def get_usage_statistics_handler(args_array):
    try:
        # If the object is usage_statistics-capable, then it will have a usage_statistics_handler
        handler = getattr(args_array[0], "_usage_statistics_handler", None)
        if handler is not None and not isinstance(handler, UsageStatisticsHandler):
            logger.debug("Invalid UsageStatisticsHandler found on object.")
            handler = None
    except IndexError:
        # A wrapped method that is not an object; this would be erroneous usage
        logger.debug(
            "usage_statistics enabled decorator should only be used on data context methods"
        )
        handler = None
    except AttributeError:
        # A wrapped method that is not usage_statistics capable
        handler = None
    except Exception as e:
        # An unknown error -- but we still fail silently
        logger.debug(
            "Unrecognized error when trying to find usage_statistics_handler: " + str(e)
        )
        handler = None
    return handler


def usage_statistics_enabled_method(
    func=None, event_name=None, args_payload_fn=None, result_payload_fn=None
):
    """
    A decorator for usage statistics which defaults to the less detailed payload schema.
    """
    if callable(func):
        if event_name is None:
            event_name = func.__name__

        @wraps(func)
        def usage_statistics_wrapped_method(*args, **kwargs):
            # if a function like `build_data_docs()` is being called as a `dry_run`
            # then we dont want to emit usage_statistics. We just return the function without sending a usage_stats message
            if "dry_run" in kwargs and kwargs["dry_run"]:
                return func(*args, **kwargs)
            # Set event_payload now so it can be updated below
            event_payload = {}
            message = {"event_payload": event_payload, "event": event_name}
            handler = None
            try:
                if args_payload_fn is not None:
                    nested_update(event_payload, args_payload_fn(*args, **kwargs))
                result = func(*args, **kwargs)
                # We try to get the handler only now, so that it *could* be initialized in func, e.g. if it is an
                # __init__ method
                handler = get_usage_statistics_handler(args)
                if result_payload_fn is not None:
                    nested_update(event_payload, result_payload_fn(result))
                message["success"] = True
                if handler is not None:
                    handler.emit(message)
            # except Exception:
            except Exception:
                message["success"] = False
                handler = get_usage_statistics_handler(args)
                if handler:
                    handler.emit(message)
                raise
            return result

        return usage_statistics_wrapped_method
    else:

        def usage_statistics_wrapped_method_partial(func):
            return usage_statistics_enabled_method(
                func,
                event_name=event_name,
                args_payload_fn=args_payload_fn,
                result_payload_fn=result_payload_fn,
            )

        return usage_statistics_wrapped_method_partial


def run_validation_operator_usage_statistics(
    data_context,  # self
    validation_operator_name,
    assets_to_validate,
    run_id=None,
    **kwargs
):
    try:
        data_context_id = data_context.data_context_id
    except AttributeError:
        data_context_id = None
    anonymizer = _anonymizers.get(data_context_id, None)
    if anonymizer is None:
        anonymizer = Anonymizer(data_context_id)
        _anonymizers[data_context_id] = anonymizer
    payload = {}
    try:
        payload["anonymized_operator_name"] = anonymizer.anonymize(
            validation_operator_name
        )
    except TypeError:
        logger.debug(
            "run_validation_operator_usage_statistics: Unable to create validation_operator_name hash"
        )
    if data_context._usage_statistics_handler:
        try:
            batch_anonymizer = data_context._usage_statistics_handler._batch_anonymizer
            payload["anonymized_batches"] = [
                batch_anonymizer.anonymize_batch_info(batch)
                for batch in assets_to_validate
            ]
        except Exception:
            logger.debug(
                "run_validation_operator_usage_statistics: Unable to create anonymized_batches payload field"
            )
    return payload


def save_expectation_suite_usage_statistics(
    data_context, expectation_suite, expectation_suite_name=None  # self
):
    try:
        data_context_id = data_context.data_context_id
    except AttributeError:
        data_context_id = None
    anonymizer = _anonymizers.get(data_context_id, None)
    if anonymizer is None:
        anonymizer = Anonymizer(data_context_id)
        _anonymizers[data_context_id] = anonymizer
    payload = {}

    if expectation_suite_name is None:
        expectation_suite_name = expectation_suite.expectation_suite_name

    try:
        payload["anonymized_expectation_suite_name"] = anonymizer.anonymize(
            expectation_suite_name
        )
    except Exception:
        logger.debug(
            "save_expectation_suite_usage_statistics: Unable to create anonymized_expectation_suite_name payload field"
        )
    return payload


def edit_expectation_suite_usage_statistics(data_context, expectation_suite_name):
    try:
        data_context_id = data_context.data_context_id
    except AttributeError:
        data_context_id = None
    anonymizer = _anonymizers.get(data_context_id, None)
    if anonymizer is None:
        anonymizer = Anonymizer(data_context_id)
        _anonymizers[data_context_id] = anonymizer
    payload = {}

    try:
        payload["anonymized_expectation_suite_name"] = anonymizer.anonymize(
            expectation_suite_name
        )
    except Exception:
        logger.debug(
            "edit_expectation_suite_usage_statistics: Unable to create anonymized_expectation_suite_name payload field"
        )
    return payload


def add_datasource_usage_statistics(data_context, name, **kwargs):
    if not data_context._usage_statistics_handler:
        return dict()
    try:
        data_context_id = data_context.data_context_id
    except AttributeError:
        data_context_id = None

    try:
        datasource_anonymizer = (
            data_context._usage_statistics_handler._datasource_anonymizer
        )
    except Exception:
        datasource_anonymizer = DatasourceAnonymizer(data_context_id)

    payload = {}
    try:
        payload = datasource_anonymizer.anonymize_datasource_info(name, kwargs)
    except Exception:
        logger.debug(
            "add_datasource_usage_statistics: Unable to create add_datasource_usage_statistics payload field"
        )
    return payload


def send_usage_message(data_context, event, event_payload=None, success=None):
    """send a usage statistics message."""
    try:
        handler = getattr(data_context, "_usage_statistics_handler", None)
        message = {
            "event": event,
            "event_payload": event_payload or {},
            "success": success,
        }
        if handler is not None:
            handler.emit(message)
    except Exception:
        pass



import copy
from typing import Dict, List, Optional, Tuple, Union

from great_expectations.core.id_dict import IDDict


class MetricConfiguration:
    def __init__(
        self,
        metric_name: str,
        metric_domain_kwargs: Dict,
        metric_value_kwargs: dict = None,
        metric_dependencies: dict = None,
    ):
        self._metric_name = metric_name
        if not isinstance(metric_domain_kwargs, IDDict):
            metric_domain_kwargs = IDDict(metric_domain_kwargs)
        self._metric_domain_kwargs = metric_domain_kwargs
        if not isinstance(metric_value_kwargs, IDDict):
            if metric_value_kwargs is None:
                metric_value_kwargs = dict()
            metric_value_kwargs = IDDict(metric_value_kwargs)
        self._metric_value_kwargs = metric_value_kwargs
        if metric_dependencies is None:
            metric_dependencies = dict()
        self.metric_dependencies = metric_dependencies

    @property
    def metric_name(self):
        return self._metric_name

    @property
    def metric_domain_kwargs(self):
        return self._metric_domain_kwargs

    @property
    def metric_value_kwargs(self):
        return self._metric_value_kwargs

    @property
    def metric_domain_kwargs_id(self):
        return self._metric_domain_kwargs.to_id()

    @property
    def metric_value_kwargs_id(self):
        return self._metric_value_kwargs.to_id()

    @property
    def id(self) -> Tuple[str, str, str]:
        return (
            self.metric_name,
            self.metric_domain_kwargs_id,
            self.metric_value_kwargs_id,
        )


class MetricEdge:
    def __init__(
        self, left: MetricConfiguration, right: Union[MetricConfiguration, None]
    ):
        self._left = left
        self._right = right

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def id(self):
        if self.right:
            return self.left.id, self.right.id
        return self.left.id, None


class ValidationGraph:
    def __init__(self, edges: Optional[List[MetricEdge]] = None):
        if edges:
            self._edges = edges
        else:
            self._edges = []

        self._edge_ids = {edge.id for edge in self._edges}

    def add(self, edge: MetricEdge):
        if edge.id not in self._edge_ids:
            self._edges.append(edge)
            self._edge_ids.add(edge.id)

    @property
    def edges(self):
        return copy.deepcopy(self._edges)



from typing import Any, Dict, Tuple

from dateutil.parser import parse

from great_expectations.execution_engine import PandasExecutionEngine
from great_expectations.execution_engine.execution_engine import (
    MetricDomainTypes,
    MetricPartialFunctionTypes,
)
from great_expectations.expectations.metrics.map_metric import MapMetricProvider
from great_expectations.expectations.metrics.metric_provider import metric_partial
from great_expectations.expectations.metrics.util import filter_pair_metric_nulls


class ColumnPairValuesAGreaterThanB(MapMetricProvider):
    condition_metric_name = "column_pair_values.a_greater_than_b"
    condition_value_keys = (
        "ignore_row_if",
        "or_equal",
        "parse_strings_as_datetimes",
        "allow_cross_type_comparisons",
    )
    domain_keys = ("batch_id", "table", "column_A", "column_B")

    @metric_partial(
        engine=PandasExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.MAP_CONDITION_SERIES,
        domain_type=MetricDomainTypes.COLUMN_PAIR,
    )
    def _pandas(
        cls,
        execution_engine: "PandasExecutionEngine",
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):

        ignore_row_if = metric_value_kwargs.get("ignore_row_if")
        if not ignore_row_if:
            ignore_row_if = "both_values_are_missing"
        or_equal = metric_value_kwargs.get("or_equal")
        parse_strings_as_datetimes = metric_value_kwargs.get(
            "parse_strings_as_datetimes"
        )
        allow_cross_type_comparisons = metric_value_kwargs.get(
            "allow_cross_type_comparisons"
        )

        df, compute_domain, accessor_domain = execution_engine.get_compute_domain(
            metric_domain_kwargs, MetricDomainTypes.COLUMN_PAIR
        )

        column_A, column_B = filter_pair_metric_nulls(
            df[metric_domain_kwargs["column_A"]],
            df[metric_domain_kwargs["column_B"]],
            ignore_row_if=ignore_row_if,
        )

        if allow_cross_type_comparisons:
            raise NotImplementedError

        if parse_strings_as_datetimes:
            temp_column_A = column_A.map(parse)
            temp_column_B = column_B.map(parse)

        else:
            temp_column_A = column_A
            temp_column_B = column_B

        if or_equal:
            return temp_column_A >= temp_column_B, compute_domain, accessor_domain
        else:
            return temp_column_A > temp_column_B, compute_domain, accessor_domain



import json

from great_expectations.core.metric import ValidationMetricIdentifier
from great_expectations.core.util import ensure_json_serializable
from great_expectations.data_context.store.database_store_backend import (
    DatabaseStoreBackend,
)
from great_expectations.data_context.store.store import Store
from great_expectations.util import load_class, verify_dynamic_loading_support


class MetricStore(Store):
    """
    A MetricStore stores ValidationMetric information to be used between runs.
    """

    _key_class = ValidationMetricIdentifier

    def __init__(self, store_backend=None, store_name=None):
        if store_backend is not None:
            store_backend_module_name = store_backend.get(
                "module_name", "great_expectations.data_context.store"
            )
            store_backend_class_name = store_backend.get(
                "class_name", "InMemoryStoreBackend"
            )
            verify_dynamic_loading_support(module_name=store_backend_module_name)
            store_backend_class = load_class(
                store_backend_class_name, store_backend_module_name
            )

            if issubclass(store_backend_class, DatabaseStoreBackend):
                # Provide defaults for this common case
                if "table_name" not in store_backend:
                    store_backend["table_name"] = store_backend.get(
                        "table_name", "ge_metrics"
                    )
                if "key_columns" not in store_backend:
                    store_backend["key_columns"] = store_backend.get(
                        "key_columns",
                        [
                            "run_name",
                            "run_time",
                            "data_asset_name",
                            "expectation_suite_identifier",
                            "metric_name",
                            "metric_kwargs_id",
                        ],
                    )

        super().__init__(store_backend=store_backend, store_name=store_name)

    # noinspection PyMethodMayBeStatic
    def _validate_value(self, value):
        # Values must be json serializable since they must be inputs to expectation configurations
        ensure_json_serializable(value)

    def serialize(self, key, value):
        return json.dumps({"value": value})

    def deserialize(self, key, value):
        if value:
            return json.loads(value)["value"]


class EvaluationParameterStore(MetricStore):
    def __init__(self, store_backend=None, store_name=None):
        if store_backend is not None:
            store_backend_module_name = store_backend.get(
                "module_name", "great_expectations.data_context.store"
            )
            store_backend_class_name = store_backend.get(
                "class_name", "InMemoryStoreBackend"
            )
            verify_dynamic_loading_support(module_name=store_backend_module_name)
            store_backend_class = load_class(
                store_backend_class_name, store_backend_module_name
            )

            # Store Backend Class was loaded successfully; verify that it is of a correct subclass.
            if issubclass(store_backend_class, DatabaseStoreBackend):
                # Provide defaults for this common case
                store_backend["table_name"] = store_backend.get(
                    "table_name", "ge_evaluation_parameters"
                )
        super().__init__(store_backend=store_backend, store_name=store_name)

    def get_bind_params(self, run_id):
        params = {}
        for k in self._store_backend.list_keys(run_id.to_tuple()):
            key = self.tuple_to_key(k)
            params[key.to_evaluation_parameter_urn()] = self.get(key)
        return params



import abc
import logging
import time
import warnings
from enum import Enum
from typing import Any

from dateutil.parser import parse

from great_expectations.exceptions import GreatExpectationsError
from great_expectations.validator.validator import Validator

from ..core.expectation_suite import ExpectationSuite
from ..core.run_identifier import RunIdentifier
from ..data_asset import DataAsset
from ..dataset import Dataset

logger = logging.getLogger(__name__)


class ProfilerDataType(Enum):
    """Useful data types for building profilers."""

    INT = "int"
    FLOAT = "float"
    STRING = "string"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    UNKNOWN = "unknown"


class ProfilerCardinality(Enum):
    """Useful cardinality categories for building profilers."""

    NONE = "none"
    ONE = "one"
    TWO = "two"
    FEW = "few"
    VERY_FEW = "very few"
    MANY = "many"
    VERY_MANY = "very many"
    UNIQUE = "unique"


class ProfilerTypeMapping:
    """Useful backend type mapping for building profilers."""

    # Future support possibility: JSON (RECORD)
    # Future support possibility: BINARY (BYTES)
    INT_TYPE_NAMES = [
        "INTEGER",
        "integer",
        "int",
        "int_",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "INT",
        "TINYINT",
        "BYTEINT",
        "SMALLINT",
        "BIGINT",
        "IntegerType",
        "LongType",
        "DECIMAL",
    ]
    FLOAT_TYPE_NAMES = [
        "FLOAT",
        "DOUBLE",
        "FLOAT4",
        "FLOAT8",
        "DOUBLE_PRECISION",
        "NUMERIC",
        "FloatType",
        "DoubleType",
        "float_",
        "float16",
        "float32",
        "float64",
        "number",
    ]
    STRING_TYPE_NAMES = [
        "CHAR",
        "VARCHAR",
        "NVARCHAR",
        "TEXT",
        "STRING",
        "StringType",
        "string",
        "str",
    ]
    BOOLEAN_TYPE_NAMES = [
        "BOOLEAN",
        "boolean",
        "BOOL",
        "TINYINT",
        "BIT",
        "bool",
        "BooleanType",
    ]
    DATETIME_TYPE_NAMES = [
        "DATETIME",
        "DATE",
        "TIME",
        "TIMESTAMP",
        "DateType",
        "TimestampType",
        "datetime64",
        "Timestamp",
    ]


class Profiler(metaclass=abc.ABCMeta):
    """
    Profilers creates suites from various sources of truth.

    These sources of truth can be data or non-data sources such as DDLs.

    When implementing a Profiler ensure that you:
    - Implement a . _profile() method
    - Optionally implement .validate() method that verifies you are running on the right
     kind of object. You should raise an appropriate Exception if the object is not valid.
    """

    def __init__(self, configuration: dict = None):
        self.configuration = configuration

    def validate(self, item_to_validate: Any) -> None:
        pass

    def profile(self, item_to_profile: Any, suite_name: str = None) -> ExpectationSuite:
        self.validate(item_to_profile)
        expectation_suite = self._profile(item_to_profile, suite_name=suite_name)
        return expectation_suite

    @abc.abstractmethod
    def _profile(
        self, item_to_profile: Any, suite_name: str = None
    ) -> ExpectationSuite:
        pass


class DataAssetProfiler:
    @classmethod
    def validate(cls, data_asset):
        return isinstance(data_asset, DataAsset)


class DatasetProfiler(DataAssetProfiler):
    @classmethod
    def validate(cls, dataset):
        return isinstance(dataset, (Dataset, Validator))

    @classmethod
    def add_expectation_meta(cls, expectation):
        expectation.meta[str(cls.__name__)] = {"confidence": "very low"}
        return expectation

    @classmethod
    def add_meta(cls, expectation_suite, batch_kwargs=None):
        class_name = str(cls.__name__)
        expectation_suite.meta[class_name] = {
            "created_by": class_name,
            "created_at": time.time(),
        }

        if batch_kwargs is not None:
            expectation_suite.meta[class_name]["batch_kwargs"] = batch_kwargs

        new_expectations = [
            cls.add_expectation_meta(exp) for exp in expectation_suite.expectations
        ]
        expectation_suite.expectations = new_expectations

        if "notes" not in expectation_suite.meta:
            expectation_suite.meta["notes"] = {
                "format": "markdown",
                "content": [
                    "_To add additional notes, edit the <code>meta.notes.content</code> field in the appropriate Expectation json file._"
                    # TODO: be more helpful to the user by piping in the filename.
                    # This will require a minor refactor to make more DataContext information accessible from this method.
                    # "_To add additional notes, edit the <code>meta.notes.content</code> field in <code>expectations/mydb/default/movies/BasicDatasetProfiler.json</code>_"
                ],
            }
        return expectation_suite

    @classmethod
    def profile(
        cls,
        data_asset,
        run_id=None,
        profiler_configuration=None,
        run_name=None,
        run_time=None,
    ):
        assert not (run_id and run_name) and not (
            run_id and run_time
        ), "Please provide either a run_id or run_name and/or run_time."
        if isinstance(run_id, str) and not run_name:
            warnings.warn(
                "String run_ids will be deprecated in the future. Please provide a run_id of type "
                "RunIdentifier(run_name=None, run_time=None), or a dictionary containing run_name "
                "and run_time (both optional). Instead of providing a run_id, you may also provide"
                "run_name and run_time separately.",
                DeprecationWarning,
            )
            try:
                run_time = parse(run_id)
            except (ValueError, TypeError):
                pass
            run_id = RunIdentifier(run_name=run_id, run_time=run_time)
        elif isinstance(run_id, dict):
            run_id = RunIdentifier(**run_id)
        elif not isinstance(run_id, RunIdentifier):
            run_name = run_name or "profiling"
            run_id = RunIdentifier(run_name=run_name, run_time=run_time)

        if not cls.validate(data_asset):
            raise GreatExpectationsError("Invalid data_asset for profiler; aborting")

        expectation_suite = cls._profile(
            data_asset, configuration=profiler_configuration
        )

        batch_kwargs = data_asset.batch_kwargs
        expectation_suite = cls.add_meta(expectation_suite, batch_kwargs)
        validation_results = data_asset.validate(
            expectation_suite, run_id=run_id, result_format="SUMMARY"
        )
        expectation_suite.add_citation(
            comment=str(cls.__name__) + " added a citation based on the current batch.",
            batch_kwargs=data_asset.batch_kwargs,
            batch_markers=data_asset.batch_markers,
            batch_parameters=data_asset.batch_parameters,
        )
        return expectation_suite, validation_results

    @classmethod
    def _profile(cls, dataset, configuration=None):
        raise NotImplementedError



# Utility methods for dealing with DataConnector objects

import copy
import logging
import os
import re
import sre_constants
import sre_parse
from pathlib import Path
from typing import Any, Dict, List, Optional

import pandas as pd

import great_expectations.exceptions as ge_exceptions
from great_expectations.core.batch import BatchDefinition, BatchRequest
from great_expectations.core.id_dict import (
    PartitionDefinition,
    PartitionDefinitionSubset,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource.data_connector.sorter import Sorter
from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyBatchData,
)

logger = logging.getLogger(__name__)

try:
    import pyspark
    import pyspark.sql as pyspark_sql
except ImportError:
    pyspark = None
    pyspark_sql = None
    logger.debug(
        "Unable to load pyspark and pyspark.sql; install optional spark dependency for support."
    )


DEFAULT_DATA_ASSET_NAME: str = "DEFAULT_ASSET_NAME"


def batch_definition_matches_batch_request(
    batch_definition: BatchDefinition,
    batch_request: BatchRequest,
) -> bool:
    assert isinstance(batch_definition, BatchDefinition)
    assert isinstance(batch_request, BatchRequest)

    if batch_request.datasource_name:
        if batch_request.datasource_name != batch_definition.datasource_name:
            return False
    if batch_request.data_connector_name:
        if batch_request.data_connector_name != batch_definition.data_connector_name:
            return False
    if batch_request.data_asset_name:
        if batch_request.data_asset_name != batch_definition.data_asset_name:
            return False

    if batch_request.partition_request:
        partition_identifiers: Any = batch_request.partition_request.get(
            "partition_identifiers"
        )
        if partition_identifiers:
            if not isinstance(partition_identifiers, dict):
                return False
            for key in partition_identifiers.keys():
                if not (
                    key in batch_definition.partition_definition
                    and batch_definition.partition_definition[key]
                    == partition_identifiers[key]
                ):
                    return False
    return True


def map_data_reference_string_to_batch_definition_list_using_regex(
    datasource_name: str,
    data_connector_name: str,
    data_asset_name: str,
    data_reference: str,
    regex_pattern: str,
    group_names: List[str],
) -> Optional[List[BatchDefinition]]:
    batch_request: BatchRequest = (
        convert_data_reference_string_to_batch_request_using_regex(
            data_reference=data_reference,
            regex_pattern=regex_pattern,
            group_names=group_names,
        )
    )
    if batch_request is None:
        return None

    if data_asset_name is None:
        data_asset_name = batch_request.data_asset_name

    return [
        BatchDefinition(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            partition_definition=PartitionDefinition(batch_request.partition_request),
        )
    ]


def convert_data_reference_string_to_batch_request_using_regex(
    data_reference: str,
    regex_pattern: str,
    group_names: List[str],
) -> Optional[BatchRequest]:
    # noinspection PyUnresolvedReferences
    matches: Optional[re.Match] = re.match(regex_pattern, data_reference)
    if matches is None:
        return None
    groups: list = list(matches.groups())
    partition_definition: PartitionDefinitionSubset = PartitionDefinitionSubset(
        dict(zip(group_names, groups))
    )

    # TODO: <Alex>Accommodating "data_asset_name" inside partition_definition (e.g., via "group_names") is problematic; we need a better mechanism.</Alex>
    # TODO: <Alex>Update: Approach -- we can differentiate "convert_data_reference_string_to_batch_request_using_regex()" methods between ConfiguredAssetFilesystemDataConnector and InferredAssetFilesystemDataConnector so that PartitionDefinition never needs to include data_asset_name. (ref: https://superconductivedata.slack.com/archives/C01C0BVPL5Q/p1603843413329400?thread_ts=1603842470.326800&cid=C01C0BVPL5Q)</Alex>
    data_asset_name: str = DEFAULT_DATA_ASSET_NAME
    if "data_asset_name" in partition_definition:
        data_asset_name = partition_definition.pop("data_asset_name")
    batch_request: BatchRequest = BatchRequest(
        data_asset_name=data_asset_name,
        partition_request=partition_definition,
    )
    return batch_request


def map_batch_definition_to_data_reference_string_using_regex(
    batch_definition: BatchDefinition,
    regex_pattern: str,
    group_names: List[str],
) -> str:

    if not isinstance(batch_definition, BatchDefinition):
        raise TypeError(
            "batch_definition is not of an instance of type BatchDefinition"
        )

    data_asset_name: str = batch_definition.data_asset_name
    partition_definition: PartitionDefinition = batch_definition.partition_definition
    partition_request: dict = partition_definition
    batch_request: BatchRequest = BatchRequest(
        data_asset_name=data_asset_name,
        partition_request=partition_request,
    )
    data_reference: str = convert_batch_request_to_data_reference_string_using_regex(
        batch_request=batch_request,
        regex_pattern=regex_pattern,
        group_names=group_names,
    )
    return data_reference


# TODO: <Alex>How are we able to recover the full file path, including the file extension?  Relying on file extension being part of the regex_pattern does not work when multiple file extensions are specified as part of the regex_pattern.</Alex>
def convert_batch_request_to_data_reference_string_using_regex(
    batch_request: BatchRequest,
    regex_pattern: str,
    group_names: List[str],
) -> str:
    if not isinstance(batch_request, BatchRequest):
        raise TypeError("batch_request is not of an instance of type BatchRequest")

    template_arguments: dict = copy.deepcopy(batch_request.partition_request)
    # TODO: <Alex>How does "data_asset_name" factor in the computation of "converted_string"?  Does it have any effect?</Alex>
    if batch_request.data_asset_name is not None:
        template_arguments["data_asset_name"] = batch_request.data_asset_name

    filepath_template: str = _invert_regex_to_data_reference_template(
        regex_pattern=regex_pattern,
        group_names=group_names,
    )
    converted_string = filepath_template.format(**template_arguments)

    return converted_string


# noinspection PyUnresolvedReferences
def _invert_regex_to_data_reference_template(
    regex_pattern: str,
    group_names: List[str],
) -> str:
    """
    NOTE Abe 20201017: This method is almost certainly still brittle. I haven't exhaustively mapped the OPCODES in sre_constants
    """
    data_reference_template: str = ""
    group_name_index: int = 0

    num_groups = len(group_names)

    # print("-"*80)
    parsed_sre = sre_parse.parse(regex_pattern)
    for token, value in parsed_sre:
        if token == sre_constants.LITERAL:
            # Transcribe the character directly into the template
            data_reference_template += chr(value)

        elif token == sre_constants.SUBPATTERN:
            if not (group_name_index < num_groups):
                break
            # Replace the captured group with "{next_group_name}" in the template
            data_reference_template += "{" + group_names[group_name_index] + "}"
            group_name_index += 1

        elif token in [
            sre_constants.MAX_REPEAT,
            sre_constants.IN,
            sre_constants.BRANCH,
            sre_constants.ANY,
        ]:
            # Replace the uncaptured group a wildcard in the template
            data_reference_template += "*"

        elif token in [
            sre_constants.AT,
            sre_constants.ASSERT_NOT,
            sre_constants.ASSERT,
        ]:
            pass
        else:
            raise ValueError(
                f"Unrecognized regex token {token} in regex pattern {regex_pattern}."
            )

    # Collapse adjacent wildcards into a single wildcard
    data_reference_template: str = re.sub("\\*+", "*", data_reference_template)
    return data_reference_template


def normalize_directory_path(
    dir_path: str, root_directory_path: Optional[str] = None
) -> str:
    # If directory is a relative path, interpret it as relative to the root directory.
    if Path(dir_path).is_absolute() or root_directory_path is None:
        return dir_path
    else:
        return str(Path(root_directory_path).joinpath(dir_path))


def get_filesystem_one_level_directory_glob_path_list(
    base_directory_path: str, glob_directive: str
) -> List[str]:
    """
    List file names, relative to base_directory_path one level deep, with expansion specified by glob_directive.
    :param base_directory_path -- base directory path, relative to which file paths will be collected
    :param glob_directive -- glob expansion directive
    :returns -- list of relative file paths
    """
    globbed_paths = Path(base_directory_path).glob(glob_directive)
    path_list: List[str] = [
        os.path.relpath(str(posix_path), base_directory_path)
        for posix_path in globbed_paths
    ]
    return path_list


def list_s3_keys(
    s3, query_options: dict, iterator_dict: dict, recursive: bool = False
) -> str:
    """
    For InferredAssetS3DataConnector, we take bucket and prefix and search for files using RegEx at and below the level
    specified by that bucket and prefix.  However, for ConfiguredAssetS3DataConnector, we take bucket and prefix and
    search for files using RegEx only at the level specified by that bucket and prefix.  This restriction for the
    ConfiguredAssetS3DataConnector is needed, because paths on S3 are comprised not only the leaf file name but the
    full path that includes both the prefix and the file name.  Otherwise, in the situations where multiple data assets
    share levels of a directory tree, matching files to data assets will not be possible, due to the path ambiguity.
    :param s3: s3 client connection
    :param query_options: s3 query attributes ("Bucket", "Prefix", "Delimiter", "MaxKeys")
    :param iterator_dict: dictionary to manage "NextContinuationToken" (if "IsTruncated" is returned from S3)
    :param recursive: True for InferredAssetS3DataConnector and False for ConfiguredAssetS3DataConnector (see above)
    :return: string valued key representing file path on S3 (full prefix and leaf file name)
    """
    if iterator_dict is None:
        iterator_dict = {}

    if "continuation_token" in iterator_dict:
        query_options.update({"ContinuationToken": iterator_dict["continuation_token"]})

    logger.debug(f"Fetching objects from S3 with query options: {query_options}")

    s3_objects_info: dict = s3.list_objects_v2(**query_options)

    if not any(key in s3_objects_info for key in ["Contents", "CommonPrefixes"]):
        raise ValueError("S3 query may not have been configured correctly.")

    if "Contents" in s3_objects_info:
        keys: List[str] = [
            item["Key"] for item in s3_objects_info["Contents"] if item["Size"] > 0
        ]
        yield from keys
    if recursive and "CommonPrefixes" in s3_objects_info:
        common_prefixes: List[Dict[str, Any]] = s3_objects_info["CommonPrefixes"]
        for prefix_info in common_prefixes:
            query_options_tmp: dict = copy.deepcopy(query_options)
            query_options_tmp.update({"Prefix": prefix_info["Prefix"]})
            # Recursively fetch from updated prefix
            yield from list_s3_keys(
                s3=s3,
                query_options=query_options_tmp,
                iterator_dict={},
                recursive=recursive,
            )
    if s3_objects_info["IsTruncated"]:
        iterator_dict["continuation_token"] = s3_objects_info["NextContinuationToken"]
        # Recursively fetch more
        yield from list_s3_keys(
            s3=s3,
            query_options=query_options,
            iterator_dict=iterator_dict,
            recursive=recursive,
        )

    if "continuation_token" in iterator_dict:
        # Make sure we clear the token once we've gotten fully through
        del iterator_dict["continuation_token"]


# TODO: <Alex>We need to move sorters and _validate_sorters_configuration() to DataConnector</Alex>
# As a rule, this method should not be in "util", but in the specific high-level "DataConnector" class, where it is
# called (and declared as private in that class).  Currently, this is "FilePathDataConnector".  However, since this
# method is also used in tests, it can remain in the present "util" module (as an exception to the above stated rule).
def build_sorters_from_config(config_list: List[Dict[str, Any]]) -> Optional[dict]:
    sorter_dict: Dict[str, Sorter] = {}
    if config_list is not None:
        for sorter_config in config_list:
            # if sorters were not configured
            if sorter_config is None:
                return None
            if "name" not in sorter_config:
                raise ValueError("Sorter config should have a name")
            sorter_name: str = sorter_config["name"]
            new_sorter: Sorter = _build_sorter_from_config(sorter_config=sorter_config)
            sorter_dict[sorter_name] = new_sorter
    return sorter_dict


def _build_sorter_from_config(sorter_config: Dict[str, Any]) -> Sorter:
    """Build a Sorter using the provided configuration and return the newly-built Sorter."""
    runtime_environment: dict = {"name": sorter_config["name"]}
    sorter: Sorter = instantiate_class_from_config(
        config=sorter_config,
        runtime_environment=runtime_environment,
        config_defaults={
            "module_name": "great_expectations.datasource.data_connector.sorter"
        },
    )
    return sorter


def fetch_batch_data_as_pandas_df(batch_data):
    if isinstance(batch_data, pd.DataFrame):
        return batch_data
    if pyspark_sql and isinstance(batch_data, pyspark_sql.DataFrame):
        return batch_data.toPandas()
    if isinstance(batch_data, SqlAlchemyBatchData):
        return batch_data.head(fetch_all=True)
    raise ge_exceptions.DataConnectorError("Unknown batch_data type encountered.")



import copy
from typing import Any, Dict, Tuple

from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
)
from great_expectations.execution_engine.execution_engine import (
    MetricDomainTypes,
    MetricPartialFunctionTypes,
)
from great_expectations.expectations.metrics.import_manager import F
from great_expectations.expectations.metrics.map_metric import MapMetricProvider
from great_expectations.expectations.metrics.metric_provider import metric_partial
from great_expectations.expectations.metrics.util import filter_pair_metric_nulls


class ColumnPairValuesEqual(MapMetricProvider):
    condition_metric_name = "column_pair_values.equal"
    condition_value_keys = ("ignore_row_if",)
    domain_keys = ("batch_id", "table", "column_A", "column_B")
    default_kwarg_values = {"ignore_row_if": "both_values_are_missing"}

    @metric_partial(
        engine=PandasExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.MAP_CONDITION_SERIES,
        domain_type=MetricDomainTypes.COLUMN_PAIR,
    )
    def _pandas(
        cls,
        execution_engine: "PandasExecutionEngine",
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):
        ignore_row_if = metric_value_kwargs.get("ignore_row_if")
        if not ignore_row_if:
            ignore_row_if = "both_values_are_missing"
        (
            df,
            compute_domain_kwargs,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(
            metric_domain_kwargs, MetricDomainTypes.COLUMN_PAIR
        )

        column_A, column_B = filter_pair_metric_nulls(
            df[metric_domain_kwargs["column_A"]],
            df[metric_domain_kwargs["column_B"]],
            ignore_row_if=ignore_row_if,
        )

        return column_A == column_B, compute_domain_kwargs, accessor_domain_kwargs

    @metric_partial(
        engine=SparkDFExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.MAP_CONDITION_FN,
        domain_type=MetricDomainTypes.COLUMN_PAIR,
    )
    def _spark(
        cls,
        execution_engine: "SparkDFExecutionEngine",
        metric_domain_kwargs: Dict,
        metric_value_kwargs: Dict,
        metrics: Dict[Tuple, Any],
        runtime_configuration: Dict,
    ):
        ignore_row_if = metric_value_kwargs["ignore_row_if"]
        compute_domain_kwargs = copy.deepcopy(metric_domain_kwargs)

        if ignore_row_if == "both_values_are_missing":
            compute_domain_kwargs["row_condition"] = (
                F.col(metric_domain_kwargs["column_A"]).isNotNull()
                & F.col(metric_domain_kwargs["column_B"]).isNotNull()
            )
            compute_domain_kwargs["condition_parser"] = "spark"
        elif ignore_row_if == "either_value_is_missing":
            compute_domain_kwargs["row_condition"] = (
                F.col(metric_domain_kwargs["column_A"]).isNotNull()
                | F.col(metric_domain_kwargs["column_B"]).isNotNull()
            )
            compute_domain_kwargs["condition_parser"] = "spark"

        (
            df,
            compute_domain_kwargs,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(
            compute_domain_kwargs, MetricDomainTypes.COLUMN_PAIR
        )

        return (
            df[metric_domain_kwargs["column_A"]]
            == df[metric_domain_kwargs["column_B"]],
            compute_domain_kwargs,
            accessor_domain_kwargs,
        )



from great_expectations.render.renderer.content_block.content_block import (
    ContentBlockRenderer,
)
from great_expectations.render.types import RenderedStringTemplateContent


class ExpectationStringRenderer(ContentBlockRenderer):
    @classmethod
    def _missing_content_block_fn(
        cls,
        configuration=None,
        result=None,
        language=None,
        runtime_configuration=None,
        **kwargs,
    ):
        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": ["alert", "alert-warning"]}},
                    "string_template": {
                        "template": "$expectation_type(**$kwargs)",
                        "params": {
                            "expectation_type": configuration.expectation_type,
                            "kwargs": configuration.kwargs,
                        },
                        "styling": {
                            "params": {
                                "expectation_type": {
                                    "classes": ["badge", "badge-warning"],
                                }
                            }
                        },
                    },
                }
            )
        ]
